from ..core.results import Results
from ..core.basejob import MultiJob
from ..core.errors import PlamsError
from ..interfaces.adfsuite.adf import ADFJob
from numpy import array as npa
from pickle import dump as pickleDump
from os.path import join as osPJ
from os.path import basename as osPB

__all__ = ['VibrationsResults' ,'VibrationsJob', 'IRJob']

try:
    from ase.vibrations import Vibrations as aseVib
    from ase.vibrations import Infrared as aseIR
    from ..tools.ase import *
except ImportError:
    __all__ = []


class VibrationsResults(Results):
    """
    A Class for handling numerical Vibrational analysis Results.

    Use ``get_ASEVib()`` to access the ``ase.vibrations.Vibrations`` object. Using e.g. ``self.get_ASEVib().summary()`` the results of the ASE frequency calculation are accessible.
    """
    def get_ASEVib(self):
        """Returns a ``ase.vibrations.Vibrations`` object if available"""
        if hasattr(self, '_vib'):
            return self._vib
        else:
            raise PlamsError('WARNING: No aseVib object present, probably Job %s failed?!' % self.job.name)


class VibrationsJob(MultiJob):
    """
    Class for calculating numerical Frequencies in parallel using arbitrary interfaces.
    This is achieved using the Vibrations class from ASE.

    *   ``molecule`` -- |Molecule| object (most propably in the chosen Methods optimized state)
    *   ``settings`` -- |Settings| instance for all Single-Point jobs to be run. Don't forget reference to a restart file if you want to save a lot of computer time!
    *   ``jobType`` -- |Job| Class you want to use.
    *   ``get_gradients`` -- Function name to retrieve gradients of the |Results| object of your chosen ``jobType.results``. Must take options ``eUnit='eV'`` and ``lUnit='Angstrom'``.
    *   ``reorder`` -- Function name of ``jobType.results`` to reorder gradients to input order, set to ``None`` if not applicable.
    *   ``aseVibOpt`` -- Options for ``ase.vibrations.Vibrations.__init__``

    The ``self.__init__()`` method calls on ase.vibrations.Vibrations.run to create all displacements, then adds them as children.
    After running, the ``self.postrun()`` method reads the gradients from all single-points and saves them according to the ASE scheme.
    Finally the ASE Vibrations object is made accessible in ``self.results.get_ASEVib`` for the retrieval of the ASE results.

    """
    _result_type = VibrationsResults

    def __init__(self, molecule, settings, jobType=ADFJob, get_gradients='get_gradients', reorder='inputOrder', aseVibOpt={}):
        MultiJob.__init__(self)
        self.molecule = molecule
        self.settings = settings
        self.jobType = jobType
        self.reorder = reorder
        self.aseVibOpt = aseVibOpt
        self.vibClass = aseVib

        self.get_grad = getattr(self.jobType._result_type, get_gradients)
        if reorder != None:
            self.reorder = getattr(self.jobType._result_type, reorder)
        else:
            self.reorder = lambda x: x


    def new_children(self):
        #Create displaced molecules with ASE
        #Include our workingdir in the name, this results in the ASE .pckl files to be saved there
        #kinda hacky but works

        if len(self.children) == 0:
            add = []
            path = self.path
            self._vib = self.vibClass(toASE(self.molecule), name=osPJ(path,'plams.vib'), **self.aseVibOpt)
            for name, atoms in self._vib.iterdisplace():
                add.append(self.jobType(molecule=fromASE(atoms), name=osPB(name), settings=self.settings))
            return add

        else:
            return None


    def postrun(self):
        #get gradients and save them to the pickle files for ASE
        path = self.path
        for child in self.children:
            res = child.results
            name = child.name
            # don't rely on gradients beeing numpy arrays
            f = [ npa(vec) for vec in self.get_grad(res, eUnit='eV', lUnit='Angstrom') ]
            force = -1.0 * npa(self.reorder(res, f))
            filename = osPJ(path, name+'.pckl')
            with open(filename, 'wb') as f:
                pickleDump(force, f, protocol=2)

        self.results._vib = self._vib



class IRJob(VibrationsJob):
    """
    Subclass of |VibrationsJob| to calculate IR modes and intensities.

    Usage is the same as for the parent class, see |VibrationsJob|.

    Additional arguments:

    * ``__init__``: get_dipole_vector, must take argument ``unit='au'``
    """
    def __init__(self, molecule, settings, jobType=ADFJob, get_gradients='get_gradients', reorder='inputOrder', get_dipole_vector='get_dipole_vector', aseVibOpt={}):
        VibrationsJob.__init__(self, molecule, settings, jobType=jobType, get_gradients=get_gradients, reorder=reorder, aseVibOpt=aseVibOpt)
        self.get_dipole = getattr(self.jobType._result_type, get_dipole_vector)
        self.vibClass = aseIR



    def postrun(self):
        #get gradients and dipole and save them to the pickle files for ASE
        path = self.path
        for child in self.children:
            res = child.results
            name = child.name
            # don't rely on gradients beeing numpy arrays
            f = [ npa(vec) for vec in self.get_grad(res, eUnit='eV', lUnit='Angstrom') ]
            force = -1.0 * npa(self.reorder(res, f))
            dipole = npa(self.get_dipole(res, unit='au'))
            filename = osPJ(path, name+'.pckl')
            with open(filename, 'wb') as f:
                pickleDump(npa([force,dipole]), f, protocol=2)

        self.results._vib = self._vib



import os
import stat
import threading
import time

try:
    import dill as pickle
except ImportError:
    import pickle

from os.path import join as opj

from .errors import FileError, JobError, PlamsError, ResultsError
from .functions import config, log
from .private import sha256
from .results import Results
from .settings import Settings
from ..mol.molecule import Molecule

__all__ = ['SingleJob', 'MultiJob']


class Job:
    """General abstract class for all kind of computational tasks.

    Methods common for all kinds of jobs are gathered here. Instances of |Job| should never be created. It should not be subclassed either. If you wish to define a new type of job please subclass either |SingleJob| or |MultiJob|.

    Methods that are meant to be explicitly called by the user are |run| and occasionally :meth:`~Job.pickle`. In most cases |pickling| is done automatically, but if for some reason you wish to do it manually, you can use :meth:`~Job.pickle` method.

    Methods that can be safely overridden in subclasses are:

    *   :meth:`~Job.check`
    *   :meth:`~Job.hash` (see |RPM|)
    *   |prerun| and |postrun| (see :ref:`prerun-postrun`)

    All other methods should remain unchanged.

    Class attribute ``_result_type`` defines the type of results associated with this job. It should point to a class and it **must** be a |Results| subclass.

    Every job instance has the following attributes:

    Attributes adjusted automatically that should not be changed by the user:

    *   ``status`` -- current status of the job. Possible values: *created*, *started*, *registered*, *running*, *finished*, *crashed*, *failed*, *successful*, *copied*, *preview*.
    *   ``results`` -- a reference to a results instance. An empty instance of the type stored in ``_result_type`` is created when the job object is created.
    *   ``path`` -- an absolute path to the job folder.
    *   ``jobmanager`` -- a job manager associated with this job.
    *   ``parent`` -- a pointer to the parent job if this job is a child job of some |MultiJob|. ``None`` otherwise.

    Attributes that can be modified, but only before |run| is called:

    *   ``name`` -- the name of the job.
    *   ``settings`` -- settings of the job.
    *   ``default_settings`` -- see :ref:`default-settings`.
    *   ``depend`` -- a list of explicit dependencies.
    *   ``_dont_pickle`` -- additional list of this instance's attributes that will be removed before pickling. See |pickling| for details.

    """

    _result_type = Results

    def __init__(self, name='plamsjob', settings=None, depend=None):
        if os.path.sep in name:
            raise PlamsError('Job name cannot contain {}'.format(os.path.sep))
        self.status = 'created'
        self.results = self.__class__._result_type(self)
        self.name = name
        self.path = None
        self.jobmanager = None
        self.parent = None
        self.settings = Settings()
        self.default_settings = [config.job]
        self.depend = depend or []
        self._dont_pickle = []
        if settings is not None:
            if isinstance(settings, Settings):
                self.settings = settings.copy()
            if isinstance(settings, Job):
                self.settings = settings.settings.copy()


    #=======================================================================


    def run(self, jobrunner=None, jobmanager=None, **kwargs):
        """Run the job using *jobmanager* and *jobrunner* (or defaults, if ``None``). Other keyword arguments (*\*\*kwargs*) are stored in ``run`` branch of job's settings. Returned value is the |Results| instance associated with this job.

        .. warning::

            This method should **not** be overridden.

        .. technical::

            This method does not do too much by itself. After some initial preparation it passes control to the job runner, which decides if a new thread should be started for this job. The role of the job runner is to execute three methods that make the full job life cycle: :meth:`~Job._prepare`, :meth:`~Job._execute` and :meth:`~Job._finalize`. During :meth:`~Job._execute` the job runner is called once again to execute the runscript (only in case of |SingleJob|).
        """
        if self.status != 'created':
            raise JobError('Trying to run previously started job {}'.format(self.name))

        self.status = 'started'
        self._log_status(1)

        self.settings.run.soft_update(Settings(kwargs))

        if jobrunner is None:
            if 'default_jobrunner' in config:
                jobrunner = config.default_jobrunner
            else:
                raise PlamsError('No default jobrunner found. This probably means that PLAMS init() was not called.')
        if jobmanager is None:
            if 'default_jobmanager' in config:
                jobmanager = config.default_jobmanager
            else:
                raise PlamsError('No default jobmanager found. This probably means that PLAMS init() was not called.')

        jobrunner._run_job(self, jobmanager)
        return self.results


    def pickle(self, filename=None):
        """Pickle this instance and save to a file indicated by *filename*. If ``None``, save to ``[jobname].dill`` in the job folder."""
        filename = filename or opj(self.path, self.name+'.dill')
        with open(filename, 'wb') as f:
            try:
                pickle.dump(self, f, -1)
            except:
                log('Pickling of {} failed'.format(self.name), 1)


    def ok(self, strict=True):
        """Check if the execution of this instance was successful. If needed, wait for the job to finish and then check if the status is *successful* (or *copied*).

            If this method is called before job's |run| method, a warning is logged and the returned value is ``False``. The most likely cause of such a behavior is simply forgetting about |run| call. However, in complicated workflows executed in parallel, it can sometimes naturally happen that one thread is ahead of others and calls :meth:`~Job.ok` before some other thread has a chance to call |run|. If you're experiencing that kind of problems, please consider using ``strict=False`` to skip the |run| check.  But keep in mind that skipping that check will deadlock the current thread if |run| never gets called.

        """
        if strict and self.status == 'created': #first thing run() does is changing the status to 'started'
            log('Job {} WARNING: ok() method was called before run(). Returned value is False. Please check the documentation'.format(self.name), 3)
            return False
        self.results.wait()
        return self.status in ['successful', 'copied']


    def check(self):
        """Check if the execution of this instance was successful. Abstract method meant for internal use."""
        raise PlamsError('Trying to run an abstract method Job.check()')


    def hash(self):
        """Calculate the hash of this instance. Abstract method meant for internal use."""
        raise PlamsError('Trying to run an abstract method Job.hash()')


    def prerun(self):
        """Actions to take before the actual job execution.

        This method is initially empty, it can be defined in subclasses or directly added to either the whole class or a single instance using |binding_decorators|.
        """
        pass


    def postrun(self):
        """Actions to take just after the actual job execution.

        This method is initially empty, it can be defined in subclasses or directly added to either the whole class or a single instance using |binding_decorators|.
        """
        pass


    #=======================================================================


    def _prepare(self, jobmanager):
        """Prepare the job for execution. This method collects steps 1-7 from :ref:`job-life-cycle`. Should not be overridden. Returned value indicates if job execution should continue (|RPM| did not find this job as previously run)."""

        log('Starting {}._prepare()'.format(self.name), 7)

        log('Resolving {}.depend'.format(self.name), 7)
        if config.preview is False:
            for j in self.depend:
                j.results.wait()
        log('{}.depend resolved'.format(self.name), 7)

        jobmanager._register(self)

        log('Starting {}.prerun()'.format(self.name), 5)
        self.prerun()
        log('{}.prerun() finished'.format(self.name), 5)

        for i in reversed(self.default_settings):
            self.settings.soft_update(i)

        prev = jobmanager._check_hash(self)
        if prev is not None:
            try:
                prev.results._copy_to(self.results)
                self.status = 'copied'
            except ResultsError as re:
                log('Copying results of {} failed because of the following error: {}'.format(prev.name, str(re)), 1)
                self.status = prev.status
            if self.settings.pickle:
                self.pickle()
            self.results.finished.set()
            self.results.done.set()
            if self.parent and self in self.parent:
                self.parent._notify()
        else:
            self.status = 'running'
            log('Starting {}._get_ready()'.format(self.name), 7)
            self._get_ready()
            log('{}._get_ready() finished'.format(self.name), 7)

        log('{}._prepare() finished'.format(self.name), 7)
        self._log_status(3)
        return prev is None


    def _get_ready(self):
        """Get ready for :meth:`~Job._execute`. This is the last step before :meth:`~Job._execute` is called. Abstract method."""
        raise PlamsError('Trying to run an abstract method Job._get_ready()')


    def _execute(self, jobrunner):
        """Execute the job. Abstract method."""
        raise PlamsError('Trying to run an abstract method Job._execute()')


    def _finalize(self):
        """Gather the results of the job execution and organize them. This method collects steps 9-12 from :ref:`job-life-cycle`. Should not be overridden."""
        log('Starting {}._finalize()'.format(self.name), 7)

        if config.preview is False:
            log('Collecting results of {}'.format(self.name), 7)
            self.results.collect()
            self.results.finished.set()
            if self.status != 'crashed':
                self.status = 'finished'
                self._log_status(3)
                if self.check():
                    log('{}.check() success. Cleaning results with keep = {}'.format(self.name, self.settings.keep), 7)
                    self.results._clean(self.settings.keep)
                    log('Starting {}.postrun()'.format(self.name), 5)
                    self.postrun()
                    log('{}.postrun() finished'.format(self.name), 5)
                    self.status = 'successful'
                    log('Pickling {}'.format(self.name), 7)
                    if self.settings.pickle:
                        self.pickle()
                else:
                    log('{}.check() failed'.format(self.name), 7)
                    self.status = 'failed'
        else:
            self.status = 'preview'
            self.results.finished.set()
        self.results.done.set()

        if self.parent and self in self.parent:
            self.parent._notify()

        log('{}._finalize() finished'.format(self.name), 7)
        self._log_status(1)


    def __getstate__(self):
        """Prepare this job instance for pickling.

        Attributes ``jobmanager``, ``parent``, ``default_settings`` and ``_lock`` are removed, as well as all attributes listed in ``self._dont_pickle``.
        """
        remove = ['jobmanager', 'parent', 'default_settings', '_lock'] + self._dont_pickle
        return {k:v for k,v in self.__dict__.items() if k not in remove}


    def _log_status(self, level):
        """Log the status of this instance on a chosen log *level*. The message is uppercased to clearly stand out among other log entries."""
        log('JOB {} {}'.format(self._full_name(), self.status.upper()), level)


    def _full_name(self):
        if self.parent:
            return '/'.join([self.parent._full_name(), self.name])
        return self.name


#===========================================================================
#===========================================================================
#===========================================================================


class SingleJob(Job):
    """Abstract class representing a job consisting of a single execution of some external binary (or arbitrary shell script in general).

    In addition to constructor arguments and attributes defined by |Job|, the constructor of this class accepts the keyword argument ``molecule`` that should be a |Molecule| instance. The constructor creates a copy of the supplied |Molecule| and stores it as the ``molecule`` attribute.

    Class attribute ``_filenames`` defines default names for input, output, runscript and error files. If you wish to override this attribute it should be a dictionary with string keys ``'inp'``, ``'out'``, ``'run'``, ``'err'``. The value for each key should be a string describing corresponding file's name. Shortcut ``$JN`` can be used for job's name. The default value is defined in the following way::

        _filenames = {'inp':'$JN.in', 'run':'$JN.run', 'out':'$JN.out', 'err': '$JN.err'}

    This class defines no new methods that could be directly called in your script. Methods that can and should be overridden are |get_input| and |get_runscript|.

    """
    _filenames = {'inp':'$JN.in', 'run':'$JN.run', 'out':'$JN.out', 'err': '$JN.err'}

    def __init__(self, molecule=None, **kwargs):
        Job.__init__(self, **kwargs)
        self.molecule = molecule.copy() if isinstance(molecule, Molecule) else molecule


    def get_input(self):
        """Generate the input file. Abstract method.

        This method should return a single string with the full content of the input file. It should process information stored in the ``input`` branch of job settings and in the ``molecule`` attribute.
        """
        raise PlamsError('Trying to run an abstract method SingleJob._get_input()')


    def get_runscript(self):
        """Generate the runscript. Abstract method.

        This method should return a single string with the runscript contents. It can process information stored in ``runscript`` branch of job  settings. In general the full runscript has the following form::

            [first line defined by job.settings.runscript.shebang]

            [contents of job.settings.runscript.pre, when present]

            [value returned by get_runscript()]

            [contents of job.settings.runscript.post, when present]

        When overridden, this method should pay attention to ``.runscript.stdout_redirect`` key in job's ``settings``.
        """
        raise PlamsError('Trying to run an abstract method SingleJob._get_runscript()')


    def hash_input(self):
        """Calculate SHA256 hash of the input file."""
        return sha256(self.get_input())


    def hash_runscript(self):
        """Calculate SHA256 hash of the runscript."""
        return sha256(self.full_runscript())


    def hash(self):
        """Calculate unique hash of this instance.

        The behavior of this method is adjusted by the value of ``hashing`` key in |JobManager| settings. If no |JobManager| is yet associated with this job, default setting from ``config.jobmanager.hashing`` is used.

        Methods :meth:`~SingleJob.hash_input` and :meth:`~SingleJob.hash_runscript` are used to obtain hashes of, respectively, input and runscript.

        Currently supported values for ``hashing`` are:

        *   ``False`` or ``None`` -- returns ``None`` and disables |RPM|.
        *   ``input`` -- returns the hash of the input file.
        *   ``runscript`` -- returns the hash of the runscript.
        *   ``input+runscript`` -- returns SHA256 hash of the concatenation of **hashes** of input and runscript.
        """
        if self.jobmanager:
            mode = self.jobmanager.settings.hashing
        else:
            mode = config.jobmanager.hashing

        if not mode:
            return None
        if mode == 'input':
            return self.hash_input()
        elif mode == 'runscript':
            return self.hash_runscript()
        elif mode == 'input+runscript':
            return sha256(self.hash_input() + self.hash_runscript())
        else:
            raise PlamsError('Unsupported hashing method: {}'.format(mode))


    def check(self):
        """Check if the calculation was successful.

        This method can be overridden in concrete subclasses of |SingleJob|. It should return a boolean value. The definition here serves as a default, to prevent crashing if a subclass does not define its own :meth:`~scm.plams.core.basejob.SingleJob.check`. It always returns ``True``.

        .. warning::

            This method is meant for internal usage and **should not** be explicitly called in your script (but it can be overridden in subclasses). Manually calling :meth:`~scm.plams.core.basejob.SingleJob.check` is not thread safe. For a thread safe function to evaluate the state of your job please use :meth:`~scm.plams.core.basejob.Job.ok`

        """
        return True


    def full_runscript(self):
        """Generate the full runscript, including shebang line and contents of ``pre`` and ``post``, if any. In practice this method is just a simple wrapper around |get_runscript|.
        """
        ret = self.settings.runscript.shebang +'\n\n'
        if 'pre' in self.settings.runscript:
            ret += self.settings.runscript.pre+'\n\n'
        ret += self.get_runscript()
        if 'post' in self.settings.runscript:
            ret += self.settings.runscript.post+'\n\n'
        return ret


    def _get_ready(self):
        """Create input and runscript files in the job folder. Methods |get_input| and :meth:`full_runscript` are used for that purpose. Filenames correspond to entries in the `_filenames` attribute"""
        inpfile = opj(self.path, self._filename('inp'))
        runfile = opj(self.path, self._filename('run'))

        with open(inpfile, 'w') as inp:
            inp.write(self.get_input())

        with open(runfile, 'w') as run:
            run.write(self.full_runscript())

        os.chmod(runfile, os.stat(runfile).st_mode | stat.S_IEXEC)


    def _execute(self, jobrunner):
        """Execute previously created runscript using *jobrunner*.

        The method :meth:`~scm.plams.core.jobrunner.JobRunner.call` of *jobrunner* is used. Working directory is ``self.path``. ``self.settings.run`` is passed as ``runflags`` argument.

        If preview mode is on, this method does nothing.
        """
        log('Starting {}._execute()'.format(self.name), 7)
        if config.preview is False:
            o = self._filename('out') if not self.settings.runscript.stdout_redirect else None
            retcode = jobrunner.call(runscript=self._filename('run'), workdir=self.path, out=o, err=self._filename('err'), runflags=self.settings.run)
            if retcode != 0:
                log('WARNING: Job {} finished with nonzero return code'.format(self.name), 3)
                self.status = 'crashed'
        log('{}._execute() finished'.format(self.name), 7)


    def _filename(self, t):
        """Return filename for file of type *t*. *t* can be any key from ``_filenames`` dictionary. ``$JN`` is replaced with job name in the returned string."""
        return self._filenames[t].replace('$JN', self.name)


    @classmethod
    def load_external(cls, path, settings=None, molecule=None, finalize=False):
        """Load an external job from *path*.

        In this context an "external job" is an execution of some external binary that was not managed by PLAMS, and hence does not have a ``.dill`` file. It can also be used in situations where the execution was started with PLAMS, but the Python process was terminated before the execution finished, resulting in steps 9-12 of :ref:`job-life-cycle` not happening.

        All the files produced by your computation should be placed in one folder and *path* should be the path to this folder or a file in this folder. The name of the folder is used as a job name. Input, output, error and runscript files, if present, should have names defined in ``_filenames`` class attribute (usually ``[jobname].in``, ``[jobname].out``, ``[jobname].err`` and ``[jobname].run``). It is not required to supply all these files, but in most cases one would like to use at least the output file, in order to use methods like :meth:`~scm.plams.core.results.Results.grep_output` or :meth:`~scm.plams.core.results.Results.get_output_chunk`. 

        This method is a class method, so it is called via class object and it returns an instance of that class::

            >>> j = SingleJob.load_external(path='some/path/jobname')
            >>> type(j)
            scm.plams.core.basejob.SingleJob
            >>> a = AMSJob.load_external(path='some/path/jobname')
            >>> type(a)
            scm.plams.interfaces.adfsuite.ams.AMSJob

        You can supply |Settings| and |Molecule| instances as *settings* and *molecule* parameters, they will end up attached to the returned job instance. If you don't do this, PLAMS will try to recreate them automatically using methods :meth:`~scm.plams.core.results.Results.recreate_settings` and :meth:`~scm.plams.core.results.Results.recreate_molecule` of the corresponding |Results| subclass. If no |Settings| instance is obtained in either way, the defaults from ``config.job`` are copied.

        You can set the *finalize* parameter to ``True`` if you wish to run the whole :meth:`~Job._finalize` on the newly created job. In that case PLAMS will perform the usual :meth:`~Job.check` to determine the job status (*successful* or *failed*), followed by cleaning of the job folder (|cleaning|), |postrun| and pickling (|pickling|). If *finalize* is ``False``, the status of the returned job is *copied*.
        """

        if not os.path.isdir(path):
            raise FileError('Path {} does not exist, cannot load from it.'.format(path))

        path = os.path.abspath(path)
        jobname = os.path.basename(path)

        job = cls(name=jobname)
        job.path = path
        job.status = 'copied'
        job.results.collect()

        job._filenames = {}
        for t,name in cls._filenames.items():
            fullname = name.replace('$JN', job.name)
            if fullname in job.results.files:
                job._filenames[t] = name
            else:
                log('The default {} file {} not present in {}'.format(t, fullname, job.path), 5)

        job.settings  = settings or job.results.recreate_settings() or config.job.copy()
        job.molecule  = molecule or job.results.recreate_molecule()

        if finalize:
            job._finalize()

        return job




#===========================================================================
#===========================================================================
#===========================================================================


class MultiJob(Job):
    """Concrete class representing a job that is a container for other jobs.

    In addition to constructor arguments and attributes defined by |Job|, the constructor of this class accepts two keyword arguments:

    *   ``children`` -- iterable container with children jobs (usually a list or a dictionary).
    *   ``childrunner`` -- by default all the children jobs are run using the same |JobRunner| as the parent job. If you wish to use a different |JobRunner| for children, you can pass it using ``childrunner``.

    Values passed as ``children`` and ``childrunner`` are stored as instance attributes and can be adjusted later, but before the |run| method is called.

    This class defines no new methods that could be directly called in your script.

    When executed, a multijob runs all its children using the same |JobManager| and |JobRunner| (unless ``childrunner`` is set). If you need to specify different run flags for children you can do it by manually setting them in child job settings::

        childjob = myjob.children[0]
        childjob.settings.run.arg = 'value'

    Since the ``run`` branch of settings gets soft-updated with run flags, the value set this way is not overwritten by parent job.

    The job folder of a multijob gets cleaned independently of its children. See |cleaning| for details.

    Private attributes ``_active_children`` and ``_lock`` are essential for proper parallel execution. Please do not modify them.
    """
    def __init__(self, children=None, childrunner=None, **kwargs):
        Job.__init__(self, **kwargs)
        self.children = [] if children is None else children
        self.childrunner = childrunner
        self._active_children = 0
        self._lock =  threading.Lock()


    def new_children(self):
        """Generate new children jobs.

        This method is useful when some of children jobs are not known beforehand and need to be generated based on other children jobs, like for example in any kind of self-consistent procedure.

        The goal of this method is to produce a new portion of children jobs. Newly created jobs should be returned in a container compatible with ``self.children`` (e.g. list for list, dict for dict). No adjustment of newly created jobs' ``parent`` attribute is needed. This method **cannot** modify ``_active_children`` attribute.

        The method defined here is a default template, returning ``None``, which means no new children jobs are generated and the entire execution of the parent job consists only of running jobs initially found in ``self.children``. To modify this behavior you can override this method in a |MultiJob| subclass or you can use one of |binding_decorators|, just like with :ref:`prerun-postrun`.
        """
        return None


    def hash(self):
        """Hashing for multijobs is disabled by default. Returns ``None``."""
        return None


    def check(self):
        """Check if the execution of this instance was successful, by calling :meth:`Job.ok` of all the children jobs.
        """
        return all([child.ok() for child in self])


    def other_jobs(self):
        """Iterate through other jobs that belong to this |MultiJob|, but are not in ``children``.

        Sometimes |prerun| or |postrun| methods create and run some small jobs that don't end up in ``children`` collection, but are still considered a part of a |MultiJob| instance (their ``parent`` atribute points to the |MultiJob| and their working folder is inside MultiJob's working folder). This method provides an iterator that goes through all such jobs.

        Each attribute of this |MultiJob| that is of type |Job| and has it's parent pointing to this |MultiJob| is returned, in a random order.
        """
        for attr in self.__dict__.values():
            if isinstance(attr, Job) and ((hasattr(attr, 'parent') and attr.parent == self) or not hasattr(attr, 'parent')):
                yield attr


    def remove_child(self, job):
        """Remove *job* from children."""

        rm = None
        for i, j in (self.children.items() if isinstance(self.children, dict) else enumerate(self.children)):
            if j == job:
                rm = i
                break
        if rm is not None:
            del self.children[rm]


    def _get_ready(self):
        """Get ready for :meth:`~MultiJob._execute`. Count children jobs and set their ``parent`` attribute."""
        self._active_children = len(self.children)
        for child in self:
            child.parent = self


    def __iter__(self):
        """Iterate through ``children``. If it is a dictionary, iterate through its values."""
        if isinstance(self.children, dict):
            return iter(self.children.values())
        return iter(self.children)


    def _notify(self):
        """Notify this job that one of its children has finished.

        Decrement ``_active_children`` by one. Use ``_lock`` to ensure thread safety.
        """
        with self._lock:
            self._active_children -= 1


    def _execute(self, jobrunner):
        """Run all children from ``children``. Then use :meth:`~MultiJob.new_children` and run all jobs produced by it. Repeat this procedure until :meth:`~MultiJob.new_children` returns an empty list. Wait for all started jobs to finish."""
        log('Starting {}._execute()'.format(self.name), 7)
        jr = self.childrunner or jobrunner

        for child in self:
            child.run(jobrunner=jr, jobmanager=self.jobmanager, **self.settings.run)

        new = self.new_children()
        while new:
            with self._lock:
                self._active_children += len(new)

            if isinstance(new, dict) and isinstance(self.children, dict):
                self.children.update(new)
                it = new.values()
            elif isinstance(new, list) and isinstance(self.children, list):
                self.children += new
                it = new
            else:
                raise JobError("ERROR in job {}: 'new_children' returned a value incompatible with 'children'".format(self.name))

            for child in it:
                child.parent = self
                child.run(jobrunner=jr, jobmanager=self.jobmanager, **self.settings.run)

            new = self.new_children()

        while self._active_children > 0:
            time.sleep(config.sleepstep)
        log('{}._execute() finished'.format(self.name), 7)




__all__ = ['global_minimum']

import sys

try:
    from rdkit.Chem import AllChem, rdForceFieldHelpers
    from ..interfaces.molecule.rdkit import to_rdmol, from_rdmol
except ImportError:
    pass

from ..core.functions import init, finish


def global_minimum(mol, n_scans=1, no_h=True, no_ring=True, bond_orders=[1.0], job_type=False, path='.', **kwarg):
    """
    Find the global minimum of the ligand (RDKit UFF or user-defined PLAMS |Job|) by systematically varying dihedral angles within the molecule.

    :param |Molecule| mol: The input molecule
    :param int n_scans:: The number of times the global minimum search should be repeated
    :param bool no_h: If hydrogen-containing bonds should ignored
    :param bool no_ring: If bonds in ring systems should ignored
    :param list bond_orders: A list of accepted bond orders (floats); a bond will be ignored if its bond order is not in *bond_orders*
    :param type or bool job_type: A type object of a class derived from |Job|. Set to ''False'' to use RDKit UFF
    :param str path: The path where the PLAMS working directory will be stored
    :param dict kwarg: Keyword arguments for *job_type*
    :return |Molecule|: A copy of *mol* with a newly optimized geometry
    """
    # Creates guess bonds if no bonds are present
    if len(mol.bonds) == 0:
        mol.guess_bonds()

    # Create a list of 2-tuples (i.e. atomic indices) representing (valid) bonds within the molecule
    bond_list = find_bond(mol, no_h=no_h, no_ring=no_ring, bond_orders=bond_orders)

    # Search for the global minimum with RDKit UFF or with PLAMS at an user-defined level of theory
    if not job_type:
        if 'rdkit.Chem.AllChem' not in sys.modules or 'rdkit.Chem.rdForceFieldHelpers' not in sys.modules:
            raise ImportError('rdkit.chem module not found, aborting RDKit UFF optimization')
        if not no_ring:
            raise TypeError('no_ring=False is not supported in combination with RDKit UFF')
        if not rdForceFieldHelpers.UFFHasAllMoleculeParams(to_rdmol(mol)):
            raise ValueError('No UFF parameters found for one or more atoms')

        for i in range(n_scans):
            for bond in bond_list:
                mol = global_minimum_scan_rdkit(mol, bond)

        # Optimize the molecule even if no valid bonds are found
        if not bond_list:
            rdmol = to_rdmol(mol)
            AllChem.UFFGetMoleculeForceField(rdmol).Minimize()
            mol = from_rdmol(rdmol)

    else:
        init(path=path)
        for i in range(n_scans):
            for bond in bond_list:
                mol = global_minimum_scan_plams(mol, bond, job_type, **kwarg)

        # Optimize the molecule even if no valid bonds are found
        if not bond_list:
            job = job_type(molecule=mol, **kwarg)
            results = job.run()
            mol = results.get_main_molecule()
        finish()

    return mol


def find_bond(mol, no_h=True, no_ring=True, bond_orders=[1.0]):
    """
    Create a list of bonds. Each entry is a tuple with indices of atoms forming a dihedral.
    Consider only diherdals with axis being a single bond, so that rotation is possible.

    :param |Molecule| mol: The input molecule
    :param bool no_h: If hydrogen-containing bonds should ignored
    :param bool no_ring: If bonds in ring systems should ignored
    :param list bond_orders: A list of accepted bond orders (floats); A bond will be ignored if its bond order is not in *bond_orders*
    :return list: A list of 2-tuples containing the atomic indices of valid bonds
    """
    mol.set_atoms_id()

    # Mark atoms that can form an "axis" of a diherdal, i.e atoms with more than one (non-hydrogen) neighbor
    for atom in mol:
        neighbors = mol.neighbors(atom)
        if no_h:
            neighbors = [at for at in mol.neighbors(atom) if at.atnum != 1]
        if no_ring:
            neighbors = [mol.in_ring(at) for at in neighbors]
        atom.mark = (len(neighbors) > 1)

    # For each bond with both ends marked add one bond to the list
    ret = []
    for i, bond in enumerate(mol.bonds):
        if bond.atom1.mark and bond.atom2.mark and bond.order in bond_orders:
            at1, at2 = bond.atom1, bond.atom2
            ret.append((at1.id-1 + 1, at2.id-1 + 1))

    # Clean up the molecule
    mol.unset_atoms_id()
    for atom in mol:
        del atom.mark

    return ret


def global_minimum_scan_plams(mol, bond_tuple, job_type, **kwarg):
    """
    Optimize the molecule (A PLAMS |Job|) with 3 different values for the given dihedral angle and find the lowest energy conformer.
    The matching PLAMS |Results| object must have access to the |get_energy()| and |get_main_molecule()| functions.
    If required, functions can be added manually to a class with the |add_to_class()| function.

    :param |Molecule| mol: The input molecule
    :param tuple bond_tuple: A 2-tuple containing the atomic indices of valid bonds
    :param type job_type: A type object of a class derived from |Job|
    :param dict kwarg: Keyword arguments for *job_type*
    :return |Molecule|: A copy of *mol* with a newly optimized geometry
    """
    # Define a number of variables and create 3 copies of the ligand
    angles = (-120, 0, 120)
    mol_list = [mol.copy() for i in range(3)]
    for angle, mol in zip(angles, mol_list):
        bond = mol[bond_tuple]
        atom = mol[bond_tuple[0]]
        mol.rotate_bond(bond, atom, angle, unit='degree')

    # Optimize the geometry for all dihedral angles in angle_list
    # The geometry that yields the minimum energy is returned
    energy_list = []
    for mol in mol_list:
        job = job_type(**kwarg)
        job.molecule = mol
        results = job.run()
        energy_list.append(results.get_energy())
        mol_new = results.get_main_molecule()
        for at, at_new in zip(mol, mol_new):
            at.coords = at_new.coords
    minimum = energy_list.index(min(energy_list))
    return mol_list[minimum]


def global_minimum_scan_rdkit(mol, bond_tuple):
    """
    Optimize the molecule (RDKit UFF) with 3 different values for the given dihedral angle and find the lowest energy conformer.

    :param |Molecule| mol: The input molecule
    :param tuple bond_tuple: A 2-tuples containing the atomic indices of valid bonds
    :return |Molecule|: A copy of *mol* with a newly optimized geometry
    """
    # Define a number of variables and create 3 copies of the ligand
    uff = AllChem.UFFGetMoleculeForceField
    angles = (-120, 0, 120)
    mol_list = [mol.copy() for i in range(3)]
    for angle, mol in zip(angles, mol_list):
        bond = mol[bond_tuple]
        atom = mol[bond_tuple[0]]
        mol.rotate_bond(bond, atom, angle, unit='degree')

    # Optimize the geometry for all dihedral angles in angle_list
    # The geometry that yields the minimum energy is returned
    mol_list = [to_rdmol(mol, properties=False) for mol in mol_list]
    for rdmol in mol_list:
        uff(rdmol).Minimize()
    energy_list = [uff(rdmol).CalcEnergy() for rdmol in mol_list]
    minimum = energy_list.index(min(energy_list))
    return from_rdmol(mol_list[minimum])



import numpy as np

from subprocess import CalledProcessError

from .scmjob import SCMJob, SCMResults
from ...core.errors import ResultsError
from ...core.settings import Settings
from ...core.functions import config, log
from ...tools.units import Units
from ...tools.periodic_table import PT


__all__ = ['ADFJob', 'ADFResults']


class ADFResults(SCMResults):
    """A specialized |SCMResults| subclass for accessing the results of |ADFJob|."""
    _kfext = '.t21'
    _rename_map = {'TAPE{}'.format(i) : '$JN.t{}'.format(i) for i in range(10,100)}


    def get_properties(self):
        """get_properties()
        Return a dictionary with all the entries from ``Properties`` section in the main KF file (``$JN.t21``).
        """
        ret = {}
        for sec,var in self._kf:
            if sec == 'Properties':
                ret[var] = self.readkf(sec,var)
        return ret


    def get_main_molecule(self):
        """get_main_molecule()
        Return a |Molecule| instance based on the ``Geometry`` section in the main KF file (``$JN.t21``).

        For runs with multiple geometries (geometry optimization, transition state search, intrinsic reaction coordinate) this is the **final** geometry. In such a case, to access the initial (or any intermediate) coordinates please use :meth:`get_input_molecule` or extract coordinates from section ``History``, variables ``xyz 1``, ``xyz 2`` and so on. Mind the fact that all coordinates written by ADF to ``History`` section are in bohr and internal atom order::

            mol = results.get_molecule(section='History', variable='xyz 1', unit='bohr', internal=True)
        """
        return self.get_molecule(section='Geometry', variable='xyz InputOrder', unit='bohr')


    def get_input_molecule(self):
        """get_input_molecule()
        Return a |Molecule| instance with initial coordinates.

        All data used by this method is taken from ``$JN.t21`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        if ('History', 'nr of geometries') in self._kf:
            return self.get_molecule(section='History', variable='xyz 1', unit='bohr', internal=True)
        return self.get_main_molecule()


    def get_energy(self, unit='au'):
        """get_energy(unit='au')
        Return final bond energy, expressed in *unit*.
        """
        return self._get_single_value('Energy', 'Bond Energy', unit)


    def get_dipole_vector(self, unit='au'):
        """get_dipole_vector(unit='au')
        Return the dipole vector, expressed in *unit*.
        """
        prop = self.get_properties()
        if 'Dipole' in prop:
            return Units.convert(prop['Dipole'], 'au', unit)
        raise ResultsError("'Dipole' not present in 'Properties' section of {}".format(self._kfpath()))


    def get_gradients(self, eUnit='au', lUnit='bohr'):
        """get_gradients(eUnit='au', lUnit='bohr')
        Return the cartesian gradients from the 'Gradients_InputOrder' field of the 'GeoOpt' Section in the kf-file, expressed in given units. Returned value is a numpy array with shape (nAtoms,3).
        """
        gradients = np.array(self.readkf('GeoOpt','Gradients_InputOrder'))
        gradients.shape = (-1,3)
        gradients *= (Units.conversion_ratio('au',eUnit) / Units.conversion_ratio('bohr',lUnit))
        return gradients


    def _extract_hessian(self, section, variable, internal_order):
        """_extract_hessian(section, variable, internal_order)
        Extract Hessian from *section*/*variable* of the TAPE21 file. Reorder from internal to input order, if *internal_order* is ``True``.
        """
        hess_int = np.array(self.readkf(section, variable))
        n = int((len(hess_int)/9 + 1)**0.5)
        hess_int.shape = (3*n,3*n)
        if internal_order:
            hess_inp = np.zeros(hess_int.shape)
            mapping = self._int2inp()
            for i in range(n):
                for j in range(n):
                    ii,jj = mapping[i]-1, mapping[j]-1
                    hess_inp[3*i:3*i+3, 3*j:3*j+3] = hess_int[3*ii:3*ii+3, 3*jj:3*jj+3]
            return hess_inp
        return hess_int


    def get_hessian(self):
        """get_hessian()
        Try extracting Hessian, either analytical or numerical, whichever is present in the TAPE21 file, in the input order. Returned value is a square numpy array of size 3*nAtoms.
        """
        if ('Hessian', 'Analytical Hessian') in self._kf:
            return self._extract_hessian('Hessian', 'Analytical Hessian', True)
        if ('Freq', 'Hessian_complete') in self._kf:
            return self._extract_hessian('Freq', 'Hessian_complete', True)
        raise ResultsError('auto_hessian: Hessian does not seem to be present in t21 file.')


    def get_energy_decomposition(self, unit='au'):
        """get_energy(unit='au')
        Return a dictionary with energy decomposition terms, expressed in *unit*.

        The following keys are present in the returned dictionary: ``Electrostatic``, ``Kinetic``, ``Coulomb``, ``XC``. The sum of all the values is equal to the value returned by :meth:`get_energy`.
        Note that additional contributions might be included, those are up to now: ``Dispersion``.
        """
        ret = {}
        ret['Electrostatic'] = self._get_single_value('Energy', 'Electrostatic Energy', unit)
        ret['Kinetic'] = self._get_single_value('Energy', 'Kinetic Energy', unit)
        ret['Coulomb'] = self._get_single_value('Energy', 'Elstat Interaction', unit)
        ret['XC'] = self._get_single_value('Energy', 'XC Energy', unit)
        if ('Energy', 'Dispersion Energy') in self._kf:
            ret['Dispersion'] = self._get_single_value('Energy', 'Dispersion Energy', unit)
        return ret


    def get_frequencies(self, unit='cm^-1'):
        """get_frequencies(unit='cm^-1')
        Return a numpy array of vibrational frequencies, expressed in *unit*.
        """
        freqs = np.array(self.readkf('Freq', 'Frequencies'))
        return freqs * Units.conversion_ratio('cm^-1', unit)


    def get_timings(self):
        """get_timings()

        Return a dictionary with timing statistics of the job execution. Returned dictionary contains keys ``cpu``, ``system`` and ``elapsed``. The values are corresponding timings, expressed in seconds.
        """
        last = self.grep_output(' Total Used : ')[-1].split()
        ret = {}
        ret['elapsed'] = float(last[-1])
        ret['system'] = float(last[-3])
        ret['cpu'] = float(last[-5])
        return ret


    def _atomic_numbers_input_order(self):
        """_atomic_numbers_input_order()
        Return a list of atomic numbers, in the input order.
        """
        n = self.readkf('Geometry', 'nr of atoms')
        tmp = self.readkf('Geometry', 'atomtype').split()
        atomtypes = {i+1 : PT.get_atomic_number(tmp[i]) for i in range(len(tmp))}
        atomtype_idx = self.readkf('Geometry', 'fragment and atomtype index')[-n:]
        atnums = [atomtypes[i] for i in atomtype_idx]
        return self.to_input_order(atnums)


    def _int2inp(self):
        """_int2inp()
        Get mapping from the internal atom order to the input atom order.
        """
        aoi = self.readkf('Geometry', 'atom order index')
        n = len(aoi)//2
        return aoi[:n]


    def recreate_molecule(self):
        """Recreate the input molecule for the corresponding job based on files present in the job folder. This method is used by |load_external|.
        """
        if self._kfpresent():
            return self.get_input_molecule()
        return None


    def recreate_settings(self):
        """Recreate the input |Settings| instance for the corresponding job based on files present in the job folder. This method is used by |load_external|.
        """
        if self._kfpresent():
            if ('General', 'Input') in self._kf:
                tmp = self.readkf('General', 'Input')
                user_input = '\n'.join([tmp[i:160+i].rstrip() for i in range(0,len(tmp),160)])
            else:
                user_input = self.readkf('General', 'user input')
            try:
                from scm.input_parser import InputParser
                with InputParser() as parser:
                    inp = parser.to_settings('adf', user_input)
            except CalledProcessError:
                from scm.input_parser import convert_legacy_input
                new_input = convert_legacy_input(user_input, program='adf')
                try:
                    with InputParser() as parser:
                        inp = parser.to_settings('adf', new_input)
                except:
                    log('Failed to recreate input settings from {}'.format(self._kf.path), 5)
                    return None
            except:
                log('Failed to recreate input settings from {}'.format(self._kf.path), 5)
                return None

            s = Settings()
            s.input = inp
            del s.input.atoms
            s.soft_update(config.job)
            return s
        return None


class ADFJob(SCMJob):
    _result_type = ADFResults
    _command = 'adf'

    def _serialize_mol(self):
        for i,atom in enumerate(self.molecule):
            smb = self._atom_symbol(atom)
            suffix = ''
            if 'adf' in atom.properties and 'fragment' in atom.properties.adf:
                suffix += 'f={fragment} '
            if 'adf' in atom.properties and 'block' in atom.properties.adf:
                suffix += 'b={block}'

            self.settings.input.atoms['_'+str(i+1)] = ('{:>5}'.format(i+1)) + atom.str(symbol=smb, suffix=suffix, suffix_dict=atom.properties.adf)

    def _remove_mol(self):
        if 'atoms' in self.settings.input:
            del self.settings.input.atoms



import numpy as np

# Calcualte bond energy of He dimers for a series of bond 
# distances using ADF

# type of atoms
atom1 = 'He'
atom2 = 'He'

# interatomic distance values
dmin = 2.2
dmax = 4.2
step = 0.2

# create a list with interatomic distances
distances = np.arange(dmin,dmax,step)

# calculation parameters (single point, TZP/PBE+GrimmeD3)
sett = Settings()
sett.input.ams.task = 'SinglePoint'
sett.input.adf.basis.type = 'TZP'
sett.input.adf.xc.gga = 'PBE'
sett.input.adf.xc.dispersion = 'Grimme3'

energies = []
for d in distances:
    mol = Molecule()
    mol.add_atom(Atom(symbol=atom1, coords=(0.0, 0.0, 0.0)))
    mol.add_atom(Atom(symbol=atom2, coords=(  d, 0.0, 0.0)))
    job = AMSJob(molecule=mol, settings=sett, name=f'dist_{d:.2f}')
    job.run()
    energies.append(job.results.get_energy(unit='kcal/mol'))

# print
print('== Results ==')
print('d[A]    E[kcal/mol]')
for d,e in zip(distances, energies):
    print(f'{d:.2f}    {e:.3f}')



from ...core.basejob import SingleJob
from ...core.settings import Settings


class ORCAJob(SingleJob):
    """
    A class representing a single computational job with ORCA
    `Orca <https://orcaforum.cec.mpg.de>`
    """

    def get_input(self):
        """
        Transform all contents of ``input`` branch of  ``settings`` into string with blocks, subblocks, keys and values. The branch self.settings.input.main corresponds to the lines starting with the special character ! in the ORCA input.

        Orca *end* keyword is mandatory for only a subset of sections. For instance the following orca input shows the keywords *methods* and *basis* use of end::

            ! UKS B3LYP/G SV(P) SV/J TightSCF Direct Grid3 FinalGrid4
            %method SpecialGridAtoms 26
                    SpecialGridIntAcc 7
                    end
            %basis NewGTO 26 "CP(PPP)" end
                   NewAuxGTO 26 "TZV/J" end
                   end

        In order to specify when the *end* keyword must be used, the following syntax can be used::

            job = ORCAJob(molecule=Molecule(<Path/to/molecule>))
            job.settings.input.main = "UKS B3LYP/G SV(P) SV/J TightSCF Direct Grid3 FinalGrid4"
            job.settings.input.method.SpecialGridAtoms = 26
            job.settings.input.method.SpecialGridIntAcc = 7

            job.settings.input.basis.NewGTO._end = '26 "CP(PPP)"'
            job.settings.input.basis.NewAuxGTO._end = '26 "TZV/J"'

        """
        def get_end(s):
            if (not isinstance(s, Settings)) or ('_end' not in s):
                return s
            else:
                return '{} end'.format(s['_end'])

        def pretty_print_inner(s, indent):
            inp = ''
            for i, (key, value) in enumerate(s.items()):
                end = get_end(value)
                if i == 0:
                    inp += ' {} {}\n'.format(key, end)
                else:
                    inp += '{}{} {}\n'.format(indent, key, end)
            return inp

        def pretty_print_orca(s, indent=''):
            inp = ''
            if isinstance(s, Settings):
                for k, v in s.items():
                    if k == 'main':
                        inp += '! {}\n\n'.format(pretty_print_orca(v, indent))
                    else:
                        indent2 = (len(k) + 2) * ' '
                        if not isinstance(v, Settings):
                            block = pretty_print_orca(v)
                        else:
                            block = pretty_print_inner(v, indent2)
                        inp += '%{}{}{}end\n\n'.format(k, block, indent2)
            elif isinstance(s, list):
                for elem in s:
                    inp += '{}'.format(elem)
            else:
                inp += '{}'.format(s)
            return inp

        inp = pretty_print_orca(self.settings.input)
        inp_mol = self.print_molecule()

        return inp + inp_mol

    def print_molecule(self):
        """Print a molecule in the ORCA format."""
        mol = self.molecule
        if mol:
            if 'charge' in mol.properties and isinstance(mol.properties.charge, int):
                charge = mol.properties.charge
            else:
                charge = 0
            if 'multiplicity' in mol.properties and isinstance(mol.properties.multiplicity, int):
                multi = mol.properties.multiplicity
            else:
                multi = 1
            xyz = '\n'.join(at.str(symbol=True, space=11, decimal=5) for at in mol.atoms)
            return '* xyz {} {}\n{}\n*\n\n'.format(charge, multi, xyz)
        else:
            return ''

    def get_runscript(self):
        """Returned runscript is just one line:
        ``orca myinput.inp``
        """
        return 'orca {}'.format(self._filename('inp'))

    def check(self):
        """Look for the normal termination signal in the output."""
        s = self.results.grep_output("ORCA TERMINATED NORMALLY")
        return len(s) > 0




#!/usr/bin/env python

from ..core.settings import Settings
from ..core.errors import PlamsError
from ..tools.kftools import KFFile
from ..interfaces.adfsuite.amsanalysis import AMSAnalysisJob
from .rkffile import RKFTrajectoryFile
from .rkfhistoryfile import RKFHistoryFile

__all__ = ['Trajectory']

class Trajectory :
        """
        Class representing an AMS trajectory

        It creates molecule objects along the trajectory, and it performs analysis
        """
        def __init__ (self, filenames) :
                """
                Initiates instance of an AMS trajectory object

                * ``filenames`` -- List of filepaths of RKF trajectory files or single filepath for RKF file

                Note: The corresponding file needs to remain on disc
                """
                if isinstance(filenames,str):
                        filenames = [filenames]

                # Ceck is RKFHistoryFile needs to be used
                classes = []
                for fn in filenames :
                        if not fn.split('.')[-1] == 'rkf' :
                                raise PlamsError('Files need to be RKF')
                        kf = KFFile(fn)
                        if kf.reader is None :
                                raise PlamsError('RKF file %s not found'%(fn))
                        kf.reader._create_index()
                        if not 'History' in kf.reader._sections.keys() :
                                raise PlamsError('%s is not a trajectory file'%(fn))
                        if 'SystemVersionHistory' in kf.reader._sections.keys() :
                                classes.append(RKFHistoryFile)
                        else :
                                classes.append(RKFTrajectoryFile)

                # Store the file objects and associated data for each file
                self.files = [classes[i](fn) for i,fn in enumerate(filenames)]
                self.molecules = [rkf.get_plamsmol() for rkf in self.files]
                self.lengths = [len(rkf) for rkf in self.files]

        def __len__ (self) :
                """
                Magic method that returns the length of the trajectory
                """
                return sum(self.lengths)

        def __iter__ (self) :
                """
                Iterates over molecule objects
                """
                for irkf,rkf in enumerate(self.files) :
                        mol = self.molecules[irkf]
                        for i in range(self.lengths[irkf]) :
                                crd,cell = rkf.read_frame(i,molecule=mol)
                                yield mol.copy()

        def __getitem__ (self, s) :
                """
                Returns Molecule object

                * ``s`` -- Python slice object
                """
                if isinstance(s,int) :
                        s = slice(s)
                        start, stop, step = s.indices(len(self))
                        indices = range(stop,stop+1)
                elif isinstance(s,slice) :
                        start, stop, step = s.indices(len(self))
                        indices = range(start,stop,step)
                mols = []
                for i in indices :
                        irkf, istep = self._get_filenum_and_stepnum(i)
                        # This is always the first molecule
                        mol = self.molecules[irkf].copy()
                        crd,cell = self.files[irkf].read_frame(istep,molecule=mol)
                        mols.append(mol)
                if len(mols) == 1 :
                        mols = mols[0]
                return mols

        def run_analysis (self, settings, steprange=None) :
                """
                Calls the AMS analysis tool behind the scene

                * ``settings``  -- PLAMS Settings object
                                   Example :
                                   settings = Settings()
                                   settings.input.Task = 'AutoCorrelation' 
                                   settings.input.AutoCorrelation.Property = 'Velocities'
                                   settings.input.AutoCorrelation.MaxStep = 2000
                * ``steprange`` -- Not implemented yet

                Returns a list of AMSAnalysisPlot objects.
                Main attributes of the AMSAnalysisPlot objects are :
                        * ``name``    -- The name of the plot
                        * ``x``       -- A list of lists containing the values of the coordinate system
                                         If the coordinate system is 1D, then it is a list containing a single list of values
                                         x = [[x1,x2,x3,x4,...,xn]]
                        * ``y``       -- A list containing the function values
                        * ``write()`` -- A method returning a string containing all plot info
                """
                #First get all the indices
                s = slice(steprange)
                start,stop,step = s.indices(len(self))
                indices = range(start,stop,step)
                if steprange is not None :
                        if len(steprange) == 1 :
                                indices = range(stop,stop+1)
                stepnums = {}
                for i in indices :
                        irkf,istep = self._get_filenum_and_stepnum(i)
                        if not irkf in stepnums :
                                stepnums[irkf] = []
                        stepnums[irkf].append(istep)
                ranges = []
                for irkf,values in stepnums.items() :
                        ranges.append((values[0]+1,values[-1]+1,step))
                print ('ranges: ',ranges)
                
                # Now completel the settings object
                trajecsettings = []
                for ikf,rkf in enumerate(self.files) :
                        s = Settings()
                        s.Trajectory.KFFilename = self.files[0].file_object.path
                        s.Trajectory.Range = '%i %i %i'%(ranges[irkf][0],ranges[irkf][1],ranges[irkf][2])
                        trajecsettings.append(s)
                settings.input.TrajectoryInfo = trajecsettings

                # Run the analysis job and extract the plots
                job = AMSAnalysisJob(settings=settings)
                result = job.run()
                plots = result.get_all_plots()
                return plots

        #################
        # Private methods
        #################

        def _get_filenum_and_stepnum (self, i) :
                """
                Connects a filenumber and a stepnumber to index i
                """
                irkf,istep = (-1,-1)
                counter = 0
                for il,length in enumerate(self.lengths) :
                        if i < counter+length :
                                irkf = il
                                istep = i-counter
                                break
                        counter += length
                if irkf < 0 :
                        raise KeyError('Index out of range')
                return irkf, istep

def make_recursively_immutable (obj,verbose=False) :
        """
        Return a new object that is immutable

        Note: Any recursive objects are NOT affected. For example, if obj is an instance of Molecule,
              then this function will return an instance of _Molecule, but new_obj.atoms[0].mol is still the original Molecule.
        """
        # Find all the branches of this object tree
        branchlist = [branch for branch in get_branches(obj)]
        if verbose:
                print ('Tree: ')
                for branch in branchlist :
                        print ([n[0] for n in branch])

        # Find length of longest branch
        maxlength = max([len(branch) for branch in branchlist])
        for i in range(maxlength-1,0,-1) :
                if verbose: print ('Level: ',i)
                # Find all objects at this level, and theit parent
                for branch in branchlist :
                        if len(branch) > i :
                                current_obj = branch[i][1]
                                current_name = branch[i][0]
                                # Now make immutable. Take care to do this only once for an object
                                if current_obj.__class__.__name__[0] == '_' or isinstance(current_obj,tuple) : continue
                                current_obj = make_immutable(current_obj)
                                parent_obj = branch[i-1][1]
                                if isinstance(parent_obj,list) :
                                        parent_obj[current_name] = current_obj
                                elif isinstance(parent_obj,dict) :
                                        parent_obj[current_name] = current_obj
                                else :
                                        parent_obj.__dict__[current_name] = current_obj

        # Now do the top layer
        current_obj = make_immutable(obj, add_self=True)
        return current_obj

def make_immutable (obj, add_self=False) :
        """
        Make this object immutable

        Note: This is specifically written for a Molecule object
        """
        def trytosetattr(self, name, value):
                """
                A setattr function that does not write attributes
                """
                #raise AttributeError("Cannot change object")
                if hasattr(self, name):
                    raise AttributeError("Cannot reassign members")
                self.__dict__[name] = value

        def trytosetitem(self, name, value) :
                """
                A setitem function for a dictionary that returns an exception
                """
                raise AttributeError("Cannot change object")

        def mutable_copy(obj, atoms=None) :
                """
                Returns a copy of self, but one that is mutable
                """
                if hasattr(obj,'__dict__') :
                        if '_orig' in obj.__dict__ :
                                return obj._orig
                        else :
                                return None
                elif '_orig' in obj :
                        return obj['_orig']
                else :
                        return None

        # If the object is a built-in object, all the below does not work
        if isinstance(obj,list) :
                return (tuple(obj))
        # There will be dictionaries (Settings objects). How to make those immutable?

        # Lets get the name of the class
        classname = obj.__class__.__name__

        # Lets get all the instance variables of the input object
        attribs = []
        values = []
        if hasattr(obj,'__dict__') :
                attribs = [key for key in obj.__dict__.keys()]
                values = [obj.__dict__[key] for key in attribs]

        # Create a new class which is derived from the objects class
        ImmutableClass = type('_'+classname, (obj.__class__,), {})

        # Now create an instance of the new class with the same instance variables
        im_obj = ImmutableClass.__new__(ImmutableClass)
        if isinstance (obj, dict) :
                # This is for a dictionary or PLAMS Settings object
                for key,value in obj.items() :
                        im_obj[key] = value
        else :
                for key,value in zip(attribs,values) :
                        setattr(im_obj,key,value)

        # Add its original self, for copying purposes
        if add_self :
                im_obj._orig = obj
                setattr(ImmutableClass,'copy',mutable_copy)

        # Now set the method that makes overwriting attributes impossible
        if isinstance(obj,dict) :
                setattr(ImmutableClass,'__setitem__',trytosetitem)
        else :
                setattr(ImmutableClass,'__setattr__',trytosetattr)

        return im_obj

def get_branches (obj, subpath=None) :
        """
        Search recursively through instance variables of object until an immutable one is met
        """
        tree = dictionary_from_object(obj)
        if subpath is None :
                subpath = (('top',obj),)
        # Identify immutable objects and circular objects (they are leafs)
        prev_objects = [node[1] for node in subpath]
        im_objs = []
        for n,o in tree.items() :
                if is_immutable(o) :
                        im_objs.append(n)
                # Avoid circular crap
                elif True in [o is po for po in prev_objects] :
                        im_objs.append(n)
        # The branch ends if the tree contains only immutable objects
        if len(im_objs) == len(tree) :
                yield subpath
        else:
                for n, o in tree.items():
                        if n in im_objs : continue
                        for path in get_branches(o, subpath+((n,o),)):
                                yield path

def is_immutable (obj) :
        """
        Check if this object is immutable

        Note: This is a first incomplete implementation
        """
        typelist = [int,float,tuple,str,bool] # This does not take into account numpy integers etc
        immutable = False
        if obj is None :
                immutable = True
        for t in typelist :
                if isinstance(obj,t) :
                        immutable = True
                        break
        return immutable

def dictionary_from_object (obj) :
        """
        Get the dicrtionary of object attributes
        """
        if hasattr(obj,'__dict__') :
                tree = obj.__dict__
        elif isinstance(obj,dict) :
                tree = obj
        elif hasattr(obj,'__iter__') :
                tree = {}
                for i,v in enumerate(obj) :
                        tree[i] = v
        else :
                raise Exception('What is this object?',type(obj))
        return tree



import math
import numpy as np

from ..core.settings import Settings
from ..tools.periodic_table import PT
from ..tools.units import Units


__all__ = ['Atom']


class Atom:
    """A class representing a single atom in three dimensional space.

    An instance of this class has the following attributes:

    *   ``atnum`` -- atomic number (zero for "dummy atoms")
    *   ``coords`` -- tuple of length 3 storing spatial coordinates
    *   ``bonds`` -- list of bonds (see |Bond|) this atom is a part of
    *   ``mol`` -- |Molecule| this atom belongs to
    *   ``properties`` -- |Settings| instance storing all other information about this atom (initially it is populated with *\*\*other*)

    The above attributes can be accessed either directly or using one of the following properties:

    *   ``x``, ``y``, ``z`` -- allow to read or modify each coordinate separately
    *   ``symbol`` -- allows to read or write the atomic symbol directly. Atomic symbol is not stored as an attribute, instead of that the atomic number (``atnum``) indicates the type of atom. In fact, ``symbol`` this is just a wrapper around ``atnum`` that uses |PeriodicTable| as a translator::

            >>> a = Atom(atnum=8)
            >>> print(a.symbol)
            O
            >>> a.symbol = 'Ca'
            >>> print(a.atnum)
            20

    *   ``mass`` -- atomic mass, obtained from |PeriodicTable|, read only
    *   ``radius`` -- atomic radius, obtained from |PeriodicTable|, read only
    *   ``connectors`` -- number of connectors, obtained from |PeriodicTable|, read only

    .. note::

        When creating a new atom, its type can be chosen either by setting an atomic number or a symbol (``atnum`` and ``symbol`` constructor arguments). The ``symbol`` argument takes precedence -- if it is supplied, the ``atnum`` argument is ignored.

    Values stored in ``coords`` tuple do not necessarily have to be numeric, you can also store any string there. This might come handy for programs that allow parametrization of coordinates in the input file (to enforce some geometry constraints for example)::

            >>> a = Atom(symbol='C', coords=(1,2,3))
            >>> print(a)
                     C       1.00000       2.00000       3.00000
            >>> a.y = 'param1'
            >>> print(a)
                     C       1.00000        param1       3.00000

    However, non-numerical coordinates cannot be used together with some methods (for example :meth:`distance_to` or :meth:`translate`). An attempt to do this raises an exception.

    Internally, atomic coordinates are always expressed in angstroms. Most of methods that read or modify atomic coordinates accept a keyword argument ``unit`` allowing to choose unit in which results and/or arguments are expressed (see |Units| for details). Throughout the entire code angstrom is the default length unit. If you don't specify ``unit`` parameter in any place of your script, all the automatic unit handling described above boils down to occasional multiplication/division by 1.0.
    """
    def __init__(self, atnum=0, symbol=None, coords=None, unit='angstrom', bonds=None, mol=None, **other):
        if symbol is not None:
            self.symbol = symbol
        else:
            self.atnum = atnum
            if atnum == 0: self._dummysymbol = 'Xx'

        self.mol = mol
        self.bonds = bonds or []
        self.properties = Settings(other)

        if coords is None:
            self.coords = (0.0, 0.0, 0.0)
        elif len(coords) == 3:
            tmp = []
            for i in coords:
                try:
                    i = Units.convert(float(i), unit, 'angstrom')
                except ValueError: pass
                tmp.append(i)
            self.coords = tuple(tmp)
        else:
            raise TypeError('Atom: Invalid coordinates passed')


    def str(self, symbol=True, suffix='', suffix_dict={}, unit='angstrom', space=14, decimal=6):
        """Return a string representation of this atom.

        Returned string is a single line (no newline characters) that always contains atomic coordinates (and maybe more). Each atomic coordinate is printed using *space* characters, with *decimal* characters reserved for decimal digits. Coordinates values are expressed in *unit*.

        If *symbol* is ``True``, atomic symbol is added at the beginning of the line. If *symbol* is a string, this exact string is printed there.

        *suffix* is an arbitrary string that is appended at the end of returned line. It can contain identifiers in curly brackets (like for example ``f={fragment}``) that will be replaced by values of corresponding keys from *suffix_dict* dictionary. See :ref:`formatstrings` for details.

        Example:

            >>> a = Atom(atnum=6, coords=(1,1.5,2))
            >>> print(a.str())
                     C      1.000000      1.500000      2.000000
            >>> print(a.str(unit='bohr'))
                     C      1.889726      2.834589      3.779452
            >>> print(a.str(symbol=False))
                  1.000000      1.500000      2.000000
            >>> print(a.str(symbol='C2.13'))
                 C2.13      1.000000      1.500000      2.000000
            >>> print(a.str(suffix='protein1'))
                     C      1.000000      1.500000      2.000000 protein1
            >>> a.properties.info = 'membrane'
            >>> print(a.str(suffix='subsystem={info}', suffix_dict=a.properties))
                     C      1.000000      1.500000      2.000000 subsystem=membrane

        """
        strformat = '{:>%is}'%space
        numformat = '{:>%i.%if}'%(space,decimal)
        f = lambda x: numformat.format(Units.convert(x, 'angstrom', unit)) if isinstance(x, (int,float)) else strformat.format(str(x))
        if symbol is False:
            return ('{0} {1} {2} '+suffix).format(*map(f,self.coords), **suffix_dict)
        if symbol is True:
            symbol = self.symbol
        return ('{0:>10s} {1} {2} {3} '+suffix).format(symbol, *map(f,self.coords), **suffix_dict)


    def __str__(self):
        """Return a string representation of this atom. Simplified version of :meth:`str` to work as a magic method."""
        return self.str()


    def __iter__(self):
        """Iteration through atom yields coordinates. Thanks to that instances of |Atom| can be passed to any method requiring as an argument a point or a vector in 3D space."""
        return iter(self.coords)


    def _setx(self, value): self.coords = (value, self.coords[1], self.coords[2])
    def _sety(self, value): self.coords = (self.coords[0], value, self.coords[2])
    def _setz(self, value): self.coords = (self.coords[0], self.coords[1], value)
    def _getx(self): return self.coords[0]
    def _gety(self): return self.coords[1]
    def _getz(self): return self.coords[2]
    x = property(_getx, _setx)
    y = property(_gety, _sety)
    z = property(_getz, _setz)


    def _getsymbol(self):
        if self.atnum == 0:
            return self._dummysymbol
        else:
            return PT.get_symbol(self.atnum)
    def _setsymbol(self, symbol):
        if symbol.lower().capitalize() in PT.dummysymbols:
            self.atnum = 0
            self._dummysymbol = symbol.lower().capitalize()
        else:
            self.atnum = PT.get_atomic_number(symbol)
            self._dummysymbol = None
    symbol = property(_getsymbol, _setsymbol)

    def _getmass(self):
        return PT.get_mass(self.atnum)
    mass = property(_getmass)

    def _getradius(self):
        return PT.get_radius(self.atnum)
    radius = property(_getradius)

    def _getconnectors(self):
        return PT.get_connectors(self.atnum)
    connectors = property(_getconnectors)

    def _ismetallic(self):
        return PT.get_metallic(self.atnum)
    is_metallic = property(_ismetallic)

    def _iselectronegative(self):
        return PT.get_electronegative(self.atnum)
    is_electronegative = property(_iselectronegative)



    def translate(self, vector, unit='angstrom'):
        """Move this atom in space by *vector*, expressed in *unit*.

        *vector* should be an iterable container of length 3 (usually tuple, list or numpy array). *unit* describes unit of values stored in *vector*.

        This method requires all atomic coordinates to be numerical values, :exc:`~exceptions.TypeError` is raised otherwise.
        """
        ratio = Units.conversion_ratio(unit, 'angstrom')
        self.coords = tuple(i + j*ratio for i,j in zip(self, vector))


    def move_to(self, point, unit='angstrom'):
        """Move this atom to a given *point* in space, expressed in *unit*.

        *point* should be an iterable container of length 3 (for example: tuple, |Atom|, list, numpy array). *unit* describes unit of values stored in *point*.

        This method requires all atomic coordinates to be numerical values, :exc:`~exceptions.TypeError` is raised otherwise.
        """
        ratio = Units.conversion_ratio(unit, 'angstrom')
        self.coords = tuple(i*ratio for i in point)


    def distance_to(self, point, unit='angstrom', result_unit='angstrom'):
        """Measure the distance between this atom and *point*.

        *point* should be an iterable container of length 3 (for example: tuple, |Atom|, list, numpy array). *unit* describes unit of values stored in *point*. Returned value is expressed in *result_unit*.

        This method requires all atomic coordinates to be numerical values, :exc:`~exceptions.TypeError` is raised otherwise.
        """
        ratio = Units.conversion_ratio(unit, 'angstrom')
        res = 0.0
        for i,j in zip(self,point):
            res += (i - j*ratio)**2
        return Units.convert(math.sqrt(res), 'angstrom', result_unit)


    def vector_to(self, point, unit='angstrom', result_unit='angstrom'):
        """Calculate a vector from this atom to *point*.

        *point* should be an iterable container of length 3 (for example: tuple, |Atom|, list, numpy array). *unit* describes unit of values stored in *point*. Returned value is expressed in *result_unit*.

        This method requires all atomic coordinates to be numerical values, :exc:`~exceptions.TypeError` is raised otherwise.
        """
        ratio = Units.conversion_ratio(unit, 'angstrom')
        resultratio = Units.conversion_ratio('angstrom', result_unit)
        return tuple((i*ratio-j)*resultratio for i,j in zip(point, self))


    def angle(self, point1, point2, point1unit='angstrom', point2unit='angstrom',result_unit='radian'):
        """Calculate an angle between vectors pointing from this atom to *point1* and *point2*.

        *point1* and *point2* should be iterable containers of length 3 (for example: tuple, |Atom|, list, numpy array). Values stored in them are expressed in, respectively, *point1unit* and *point2unit*. Returned value is expressed in *result_unit*.

        This method requires all atomic coordinates to be numerical values, :exc:`~exceptions.TypeError` is raised otherwise.
        """
        num = np.dot(self.vector_to(point1, point1unit), self.vector_to(point2, point2unit))
        den = self.distance_to(point1, point1unit) * self.distance_to(point2, point2unit)
        return Units.convert(math.acos(num/den), 'radian', result_unit)


    def rotate(self, matrix):
        """Rotate this atom according to a rotation *matrix*.

        *matrix* should be a container with 9 numerical values. It can be a list (tuple, numpy array etc.) listing matrix elements row-wise, either flat (``[1,2,3,4,5,6,7,8,9]``) or in two-level fashion (``[[1,2,3],[4,5,6],[7,8,9]]``).

        .. note::

            This method does not check if *matrix* is a proper rotation matrix.
        """
        matrix = np.array(matrix).reshape(3,3)
        self.coords = tuple(np.dot(matrix, np.array(self.coords)))

    def neighbors(self):
        """Return a list of neighbors of this atom within the molecule. The list follows the same order as the ``bonds`` attribute."""
        return [b.other_end(self) for b in self.bonds]



import os
from scm.plams import Settings, JobError, AMSJob, CRSJob, Molecule, AMSResults, CRSResults, KFFile

__all__ = ['run_crs_ams']


def run_crs_ams(settings_ams, settings_crs,
                solvents, solutes=None,
                return_amsresults=False, **kwargs):
    """A workflow for running COSMO-RS calculations with AMS (*i.e.* DFT) COSMO surface charges.

    The workflow consists of four distinct steps:

    1. Perform gas-phase |AMSJob| calculations on the solvents and solutes (see *settings_ams*).
    2. Perform COSMO |AMSJob| calculations using the .rkf file from step 1
    3. Extract the relevant COSMO information from the results and create .coskf files.
    4. Perform a COSMO-RS calculations with the .coskf files produced in steps 2 and 3
       (see *settings_crs*).
       This calculation is conducted for all possible solvent/solute pairs,
       assuming solutes have been specified by the user.

    The ams solvation block (*ams_settings.input.solvation*) is soft updated with suitable
    settings for constructing COSMO-RS compatible surface charges
    (see :func:`.add_solvation_block`).
    No user-specified values are overwritten during this process.

    .. admonition:: Examples

        An example value for *settings_ams*:

        .. code:: python

            >>> from scm.plams import Settings

            >>> settings_ams = Settings()
            >>> settings_ams.input.basis.type = 'TZ2P'
            >>> settings_ams.input.xc.gga = 'BP86'
            >>> settings_ams.input.scf.converge = '1.0e-06'

        An example value for *settings_crs* (`activity coefficient`_ calculation):

        .. code:: python

            >>> settings_crs = Settings()
            >>> settings_crs.input.temperature = 298.15
            >>> settings_crs.input.property._h = 'activitycoef'

        And finally the actual calculation with methanol, ethanol and propanol as solvents and
        acetic acid as solute:

        .. code:: python

            >>> solvents = [Molecule('methanol.xyz'), Molecule('ethanol.xyz'), Molecule('propanol.xyz')]
            >>> solutes = Molecule('acetic_acid.xyz')

            >>> crs_dict = run_crs_ams(settings_ams, settings_crs, solvents, solutes)
            >>> for key in crs_dict:
            >>>     print("\\nSystem:", key) 
            >>>     res = crs_dict[key].get_results(>>>settings_crs.input.property._h.upper())
            >>>     print("activity coefficients:")
            >>>     print(res["gamma"])

            System: CRSJob.methanol.acetic_acid
            activity coefficients:
            [[1.        ]
             [0.48067852]]

            System: CRSJob.ethanol.acetic_acid
            activity coefficients:
            [[1.        ]
             [0.43502696]]

            System: CRSJob.propanol.acetic_acid
            activity coefficients:
            [[1.        ]
             [0.58399355]]

    :type settings_ams: :class:`.Settings`
    :parameter settings_ams:
        A Settings instance with settings for :class:`.AMSJob` (see Examples).

    :type settings_crs: :class:`.Settings`
    :parameter settings_crs:
        A Settings instance with settings for :class:`.CRSJob` (see Examples).

    :type solvents: |Molecule| or :class:`list` [|Molecule|]
    :parameter solvents: A Molecule or list of one or more Molecules representing solvents.

    :type solutes: |Molecule| or :class:`list` [|Molecule|], optional
    :parameter solutes: An optional Molecule or list of one or more Molecules representing solutes.

    :type return_amsresults: :class:`bool`
    :parameter return_amsresults: If ``True``, return both the solvent and solute AMS results in addition to the final COSMO-RS.

    :parameter \**kwargs, optional:
        Optional keyword arguments that will be passed to all calls of :meth:`.Job.run`.
        For example, one could consider passing a custom :ref:`job_runners` or :ref:`job_manager`.

    :returns: A dictionary with the resulting COSMO-RS output.
        The `name` of each :class:`.CRSResults` instance is used as key.
        If ``return_amsresults=True``, return both the COSMO-RS and AMS solvent and solute results.
    :rtype: :class:`dict` or :class:`tuple` [:class:`dict`, :class:`list`, :class:`list`]

    .. _`activity coefficient`: ../../COSMO-RS/Properties.html#activity-coefficients-solvent-and-solute
    """  

    solvents = [solvents] if isinstance(solvents, Molecule) else solvents
    solutes = [solutes] if isinstance(solutes, Molecule) else solutes

    # soft update to defaults
    _settings_ams = update_adf_defaults(settings_ams)
    _settings_ams = add_solvation_block(_settings_ams)
    # Validate arguments
    _validate_settings_ams(_settings_ams)

    # Decapitalize the "solvation" and "allpoints" keys
    if 'solvation' in _settings_ams.settings.input.adf:
        _settings_ams.input.adf.solvation = _settings_ams.settings.input.adf.pop('solvation')

    # Create the COSMO surfaces for the solute
    solvent_list = [run_amsjob(mol, _settings_ams, **kwargs) for mol in solvents]
    solute_list = [run_amsjob(mol, _settings_ams, **kwargs) for mol in solutes]
    if not solutes:
        solute_list = [None]

    # Start the and return the COSMO-RS job
    crs = {}
    for solvent in solvent_list:
        for solute in solute_list:
            results = run_crsjob(solvent, settings_crs, solute=solute, **kwargs)
            crs[results.job.name] = results

    if not return_amsresults:
        return crs
    else:
        return crs, solvent_list, solute_list


def run_amsjob(mol: Molecule, s: Settings, **kwargs) -> AMSResults:
    """Run an :class:`.AMSJob` on *mol* using the settings provided in *s*."""
    name = 'AMSJob.' + mol.properties.name if 'name' in mol.properties else 'AMSJob.mol'

    # Do a geometry optimization in the gas phase
    _s = s.copy()
    _s.input.ams.soft_update({"Task":"GeometryOptimization"})
    job1 = AMSJob(molecule=mol, settings=_s, name=name+'.gas')
    del job1.settings.input.adf.solvation
    results1 = job1.run(**kwargs)
    ams_out = results1['ams.rkf']
    adf_out = results1['adf.rkf']

    # Construct the AMS COSMO surface
    _s.input.ams.Task = "SinglePoint"
    _s.input.ams.EngineRestart = adf_out
    _s.input.ams.LoadSystem.File = ams_out
    job2 = AMSJob(molecule=None, settings=_s, depend=[job1], name=name)
    results2 = job2.run(**kwargs)

    coskf_name = job2.name.split('.')[1] 
    convert_to_coskf(results2, coskf_name)
    return results2

def convert_to_coskf(res: AMSResults, name: str):

    f = KFFile(res['adf.rkf'])
    cosmo = f.read_section("COSMO")
    # print (os.path.dirname(res.rkfpath(file='adf')), name)
    coskf = KFFile(os.path.join(os.path.dirname(res.rkfpath(file='adf')),name+".coskf"))
    for k,v in cosmo.items():
        coskf.write("COSMO",k,v)
    res.collect()

def run_crsjob(solvent: AMSResults, s: Settings, solute: AMSResults = None, **kwargs) -> CRSResults:
    """Run an :class:`.CRSJob` on with *solvent* and, optionally, *solute* using the settings provided in *s*."""
    name = 'CRSJob.' + solvent.job.name.split('.')[1]

    solv_coskf = solvent.job.name.split('.')[1]+".coskf"
    if solute is not None:
        name += '.' + solute.job.name.split('.')[1]
        solute_coskf = solute.job.name.split('.')[1]+".coskf"
        job = CRSJob(settings=s, depend=[solvent.job, solute.job], name=name)
        set_header(job.settings, solvent[solv_coskf], solute[solute_coskf])
    else:
        job = CRSJob(settings=s, depend=[solvent.job], name=name)
        set_header(job.settings, solvent[solv_coskf])

    return job.run(**kwargs)


def set_header(s: Settings, *values: str) -> None:
    """Assign *value* to the ``["_h"]`` key in *s.input.compound*."""
    s.input.compound = []
    for item in values:
        s.input.compound.append(Settings({'_h': item}))
    s.input.compound[0].frac1 = 1.0  # The first item in *values should be the solvent


def add_solvation_block(ams_settings: Settings) -> None:
    """Add the solvation block to *ams_settings*, returning a copy of the new settings.

    The solvation block (*ams_settings.input.solvation*) is soft updated
    with the following Settings:

    .. code::

        c-mat:  Exact
        charged:        method=Conj
        radii:
            Br:       2.16
            C:        2.0
            Cl:       2.05
            F:        1.72
            H:        1.3
            I:        2.32
            N:        1.83
            O:        1.72
            P:        2.13
            S:        2.16
            Si:       2.48
        scf:    Var All
        solv:   name=CRS cav0=0.0 cav1=0.0
        surf:   Delley

    """

    # Find the solvation key
    solvation = ams_settings.input.adf.find_case('solvation')
    solvation_block = ams_settings.input.adf[solvation]

    # Find all keys for within the solvation block
    keys = ('surf', 'solv', 'charged', 'c-mat', 'scf', 'radii')
    surf, solv, charged, cmat, scf, radii = [solvation_block.find_case(item) for item in keys]

    # Construct the default solvation block
    solvation_block_new = {
        surf: 'Delley',
        solv: 'name=CRS cav0=0.0 cav1=0.0',
        charged: 'method=Conj',
        cmat: 'Exact',
        scf: 'Var All',
        radii: {
            'H': 1.30,
            'C': 2.00,
            'N': 1.83,
            'O': 1.72,
            'F': 1.72,
            'Si': 2.48,
            'P': 2.13,
            'S': 2.16,
            'Cl': 2.05,
            'Br': 2.16,
            'I': 2.32
        }
    }

    # Copy ams_settings and perform a soft update
    ret = ams_settings.copy()
    ret.input.adf[solvation].soft_update(solvation_block_new)
    return ret

def update_adf_defaults(ams_settings: Settings) -> None:
    """Add the COSMO-RS compound defaults to *ams_settings*, returning a copy of the new settings.

    The engine block (*ams_settings.input.adf*) is soft updated
    with the following Settings:

    .. code::

        Engine ADF
            Basis
                Type TZP
                Core Small
            End
            XC
                GGA BP86
            End
            Relativity
                Level Scalar
            End
            BeckeGrid
                Quality Good
            End
        EndEngine

    """

    # Find the solvation key
    # solvation = ams_settings.input.adf.find_case('solvation')
    # solvation_block = ams_settings.input.adf[solvation]

    adf = ams_settings.input.find_case('adf')
    adf_block = ams_settings.input[adf]

    # Find all keys for within the adf block
    keys = ('basis', 'xc', 'relativity', 'BeckeGrid')
    basis, xc, relativity, BeckeGrid = [adf_block.find_case(item) for item in keys]

    # Construct the default solvation block
    adf_defaults_block = Settings({
        basis: {
                'Type':'TZP',
                'Core':'Small'
                },
        xc:         {'GGA':'BP86'},
        relativity: {'Level':'Scalar'},
        BeckeGrid:  {'Quality':'Good'}
    })
    # Copy ams_settings and perform a soft update
    ret = ams_settings.copy()
    ret.input.adf.soft_update(adf_defaults_block)
    return ret


def _validate_settings_ams(s: Settings) -> None:
    """Validate the *settings_ams* argument in :func:`run_crs_ams`."""
    solvation = s.input.adf.find_case('solvation')
    solv = s.input.adf[solvation].find_case('solv')

    if solvation not in s.input.adf:
        raise JobError("run_crs_ams: The 'solvation' key is absent"
                       " from settings_ams.input'")

    if solv not in s.input.adf[solvation]:
        raise JobError("run_crs_ams: The 'solv' key is absent"
                       " from settings_ams.input.solvation")

    if 'name=crs' not in s.input.adf[solvation][solv].lower():
        raise JobError("run_crs_ams: The 'name=CRS' value is absent"
                       " from settings_ams.input.solvation.solv")



import os
import re
import shutil
import sys
import threading
import time
import types
import warnings
from typing import Callable, Dict, NoReturn

from os.path import join as opj
from os.path import isfile, isdir, expandvars, dirname

from .errors import PlamsError, FileError
from .settings import Settings

__all__ = ['init', 'finish', 'log', 'load', 'load_all', 'delete_job', 'add_to_class', 'add_to_instance', 'config', 'read_molecules', 'read_all_molecules_in_xyz_file']

config = Settings()
config.init = False

#===========================================================================


def init(path=None, folder=None, config_settings:Dict=None):
    """Initialize PLAMS environment. Create global ``config`` and the default |JobManager|.

    An empty |Settings| instance is created and populated with default settings by executing ``plams_defaults``. The following locations are used to search for the defaults file, in order of precedence:

    *   If ``$PLAMSDEFAULTS`` variable is in your environment and it points to a file, this file is used (executed as a Python script).
    *   If ``$AMSHOME`` variable is in your environment and ``$AMSHOME/scripting/scm/plams/plams_defaults`` exists, it is used.
    *   Otherwise, the path ``../plams_defaults`` relative to the current file (``functions.py``) is checked. If defaults file is not found there, an exception is raised.

    Then a |JobManager| instance is created as ``config.default_jobmanager`` using *path* and *folder* to determine the main working folder. Settings for this instance are taken from ``config.jobmanager``. If *path* is not supplied, the current directory is used. If *folder* is not supplied, ``plams_workdir`` is used.

    Optionally, an additional `dict` (or |Settings| instance) can be provided to the `config_settings` argument which will be used to update the values from the ``plams_defaults``.

    .. warning::
      This function **must** be called before any other PLAMS command can be executed. Trying to do anything without it results in a crash. See also |master-script|.
    """

    if config.init == True:
        return

    if 'PLAMSDEFAULTS' in os.environ and isfile(expandvars('$PLAMSDEFAULTS')):
        defaults = expandvars('$PLAMSDEFAULTS')
    elif 'AMSHOME' in os.environ and isfile(opj(expandvars('$AMSHOME'), 'scripting', 'scm', 'plams', 'plams_defaults')):
        defaults = opj(expandvars('$AMSHOME'), 'scripting', 'scm', 'plams', 'plams_defaults')
    else:
        defaults = opj(dirname(dirname(__file__)), 'plams_defaults')
        if not isfile(defaults):
            raise PlamsError('plams_defaults not found, please set PLAMSDEFAULTS or AMSHOME in your environment')
    with open(defaults, 'r') as f:
        exec(compile(f.read(), defaults, 'exec'))

    config.update(config_settings or {})

    from .jobmanager import JobManager
    config.default_jobmanager = JobManager(config.jobmanager, path, folder)

    log('Running PLAMS located in {}'.format(dirname(dirname(__file__))), 5)
    log('Using Python {}.{}.{} located in {}'.format(*sys.version_info[:3], sys.executable), 5)
    log('PLAMS defaults were loaded from {}'.format(defaults), 5)

    log('PLAMS environment initialized', 5)
    log('PLAMS working folder: {}'.format(config.default_jobmanager.workdir), 1)

    try:
        import dill
    except ImportError:
        log('WARNING: importing dill package failed. Falling back to the default pickle module. Expect problems with pickling', 1)

    config.init = True


#===========================================================================


def finish(otherJM=None):
    """Wait for all threads to finish and clean the environment.

    This function must be called at the end of your script for |cleaning| to take place. See |master-script| for details.

    If you used some other job managers than just the default one, they need to be passed as *otherJM* list.
    """
    if config.init == False:
        return

    for thread in threading.enumerate():
        if thread.name == 'plamsthread':
            thread.join()

    config.default_jobmanager._clean()
    if otherJM:
        for jm in otherJM:
            jm._clean()
    log('PLAMS environment cleaned up successfully', 5)
    log('PLAMS run finished. Goodbye', 3)

    if config.erase_workdir is True:
        shutil.rmtree(config.default_jobmanager.workdir)

    config.init = False


#===========================================================================


def load(filename):
    """Load previously saved job from ``.dill`` file. This is just a shortcut for |load_job| method of the default |JobManager| ``config.default_jobmanager``."""
    return config.default_jobmanager.load_job(filename)


#===========================================================================


def load_all(path, jobmanager=None):
    """Load all jobs from *path*.

    This function works as multiple executions of |load_job|. It searches for ``.dill`` files inside the directory given by *path*, yet not directly in it, but one level deeper. In other words, all files matching ``path/*/*.dill`` are used. That way a path to the main working folder of a previously run script can be used to import all the jobs run by that script.

    In case of partially failed |MultiJob| instances (some children jobs finished successfully, but not all) the function will search for ``.dill`` files in children folders. That means, if ``path/[multijobname]/`` contains some subfolders (for children jobs) but does not contail a ``.dill`` file (the |MultiJob| was not fully successful), it will look into these subfolders. This behavior is recursive up to any folder tree depth.

    The purpose of this function is to provide a quick way of restarting a script. Loading all successful jobs from the previous run prevents double work and allows the new execution of the script to proceed directly to the place where the previous execution failed.

    Jobs are loaded using default job manager stored in ``config.default_jobmanager``. If you wish to use a different one you can pass it as *jobmanager* argument of this function.

    Returned value is a dictionary containing all loaded jobs as values and absolute paths to ``.dill`` files as keys.
    """
    jm = jobmanager or config.default_jobmanager
    loaded_jobs = {}
    for foldername in filter(lambda x: isdir(opj(path,x)), os.listdir(path)):
        maybedill = opj(path,foldername,foldername+'.dill')
        if isfile(maybedill):
            job = jm.load_job(maybedill)
            if job:
                loaded_jobs[os.path.abspath(maybedill)] = job
        else:
            loaded_jobs.update(load_all(path=opj(path,foldername), jobmanager=jm))
    return loaded_jobs


#===========================================================================


def delete_job(job):
    """Remove *job* from its corresponding |JobManager| and delete the job folder from the disk. Mark *job* as 'deleted'."""

    if job.status != 'created':
        job.results.wait()

    #In case job.jobmanager is None, run() method was not called yet, so no JobManager knows about this job and no folder exists.
    if job.jobmanager is not None:
        job.jobmanager.remove_job(job)
        shutil.rmtree(job.path)

    if job.parent is not None:
        job.parent.remove_child(job)

    job.status = 'deleted'
    job._log_status(5)


#===========================================================================


def read_molecules(folder, formats=None):
    """Read all molecules from *folder*.

    Read all the files present in *folder* with extensions compatible with :meth:`Molecule.read<scm.plams.mol.molecule.Molecule.read>`. Returned value is a dictionary with keys being molecule names (filename without extension) and values being |Molecule| instances.

    The optional argument *formats* can be used to narrow down the search to files with specified extensions::

        molecules = read_molecules('mymols', formats=['xyz', 'pdb'])
    """
    from ..mol.molecule import Molecule
    extensions = formats or list(Molecule._readformat.keys())
    is_valid = lambda x: isfile(opj(folder,x)) and any([x.endswith('.'+ext) for ext in extensions])
    filenames = filter(is_valid, os.listdir(folder))
    ret = {}
    for f in filenames:
        m = Molecule(opj(folder,f))
        ret[m.properties.name] = m
    return ret


#===========================================================================


def read_all_molecules_in_xyz_file(filename):
    """The .xyz format allows to store multiple geometries on a single file (such file is essentially a concatenated series of xyz files)
    
    This function returns a *list* of all molecules found in the file *filename*

    *filename*: path (absolute or relative to the current working directory) to the xyz file.
    """
    from ..mol.molecule import Molecule
    mols = []
    with open(filename, 'r') as f:
        while True:
            try:
                mol = Molecule()
                mol.readxyz(f)
                mols.append(mol)
            except FileError:
                break
    return mols


#===========================================================================


_stdlock = threading.Lock()
_filelock = threading.Lock()

def log(message, level=0):
    """Log *message* with verbosity *level*.

    Logs are printed independently to the text logfile (a file called ``logfile`` in the main working folder) and to the standard output. If *level* is equal or lower than verbosity (defined by ``config.log.file`` or ``config.log.stdout``) the message is printed. Date and/or time can be added based on ``config.log.date`` and ``config.log.time``. All logging activity is thread safe.
    """
    if 'log' in config:
        if level <= config.log.file or level <= config.log.stdout:
            message = str(message)
            prefix = ''
            if config.log.date:
                prefix += '%d.%m|'
            if config.log.time:
                prefix += '%H:%M:%S'
            if prefix:
                prefix = '[' + prefix.rstrip('|') + '] '
                message = time.strftime(prefix) + message
            if level <= config.log.stdout:
                with _stdlock:
                    print(message)
            if level <= config.log.file and 'default_jobmanager' in config:
                with _filelock, open(config.default_jobmanager.logfile, 'a') as f:
                    f.write(message + '\n')


#===========================================================================


def add_to_class(classname):
    """Add decorated function as a method to the whole class *classname*.

    The decorated function should follow a method-like syntax, with the first argument ``self`` that references the class instance.
    Example usage::

        @add_to_class(ADFResults)
        def get_energy(self):
            return self.readkf('Energy', 'Bond Energy')

    After executing the above code all instances of ``ADFResults`` in the current script (even the ones created beforehand) are enriched with ``get_energy`` method that can be invoked by::

        someadfresults.get_energy()

    The added method is accessible also from subclasses of *classname* so ``@add_to_class(Results)`` in the above example will work too.

    If *classname* is |Results| or any of its subclasses, the added method will be wrapped with the thread safety guard (see |parallel|).
    """
    from .results import _restrict, _MetaResults
    def decorator(func):
        if isinstance(classname, _MetaResults):
            func = _restrict(func)
        setattr(classname, func.__name__, func)
    return decorator


#===========================================================================


def add_to_instance(instance):
    """Add decorated function as a method to one particular *instance*.

    The decorated function should follow a method-like syntax, with the first argument ``self`` that references the class instance.
    Example usage::

        results = myjob.run()

        @add_to_instance(results)
        def get_energy(self):
            return self.readkf('Energy', 'Bond Energy')

        results.get_energy()

    The added method is accessible only for that one particular instance and it overrides any methods with the same name defined on a class level (in original class' source) or added with :func:`add_to_class` decorator.

    If *instance* is an instance of |Results| or any of its subclasses, the added method will be wrapped with the thread safety guard (see |parallel|).
    """
    from .results import _restrict, Results
    def decorator(func):
        if isinstance(instance, Results):
            func = _restrict(func)
        func = types.MethodType(func, instance)
        setattr(instance, func.__func__.__name__, func)
    return decorator


#===========================================================================


def parse_heredoc(bash_input: str, heredoc_delimit: str = 'eor') -> str:
    """Take a string and isolate the content of a bash-style `Here Document`_.

    The input string, *bash_input*, is returned unaltered if no heredoc block is found.
    If multiple heredoc blocks are present only the first one is returned.

    An example bash input file for ADF:

    .. code:: bash

        #!/bin/bash

        $AMSBIN/adf << eor
        ATOMS
            1.H  0.0  0.0  0.0
            2.H  1.0  0.0  0.0
        END

        BASIS
            type TZ2P
        END

        XC
            GGA BP86
        END
        eor

        echo "Job finished"

    The matching :func:`parse_heredoc` output:

    .. code:: python

        >>> filename: str = ...  # The bash input file
        >>> with open(filename, 'r') as f:
        ...     output = parse_heredoc(f.read())

        >>> print(output)
        ATOMS
            1.H  0.0  0.0  0.0
            2.H  1.0  0.0  0.0
        END

        BASIS
            type TZ2P
        END

        XC
            GGA BP86
        END

    .. _`Here Document`: https://en.wikipedia.org/wiki/Here_document

    """
    # Find the start of the heredoc block
    start_pattern = r'<<(-)?(\s+)?{}'.format(heredoc_delimit)
    start_heredoc = re.search(start_pattern, bash_input)
    if not start_heredoc:
        return bash_input

    # Find the end of the heredoc block
    end_pattern = r'\n(\s+)?{}(\s+)?\n'.format(heredoc_delimit)
    end_heredoc = re.search(end_pattern, bash_input)

    # Prepare the slices
    try:
        i, j = start_heredoc.end(), end_heredoc.start()
    except AttributeError as ex:
        err = f"parse_heredoc: failed to find the final '{heredoc_delimit}' delimiter"
        raise ValueError(err).with_traceback(ex.__traceback__)

    # Grab heredoced block and parse it
    _, ret = bash_input[i:j].split('\n', maxsplit=1)
    return ret



import os
import shutil
import signal
import subprocess
import time
import threading
import queue
import struct
import weakref
import tempfile
import functools
import numpy as np
import collections
from typing import *

TMPDIR = os.environ['SCM_TMPDIR'] if 'SCM_TMPDIR' in os.environ else None

if os.name == 'nt':
    import ctypes
    import ctypes.wintypes
    import msvcrt

    kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)

    def CheckHandle(result, func, arguments):
        if result == ctypes.wintypes.HANDLE(-1).value:
            raise ctypes.WinError(ctypes.get_last_error())
        else:
            return result

    CreateNamedPipe = kernel32.CreateNamedPipeW
    CreateNamedPipe.restype = ctypes.wintypes.HANDLE
    CreateNamedPipe.argtypes = [ctypes.c_wchar_p] + [ctypes.wintypes.DWORD]*6 + [ctypes.c_void_p]
    CreateNamedPipe.errcheck = CheckHandle

    PIPE_REJECT_REMOTE_CLIENTS = 0x00000008
    PIPE_ACCESS_DUPLEX = 0x00000003
    PIPE_TYPE_BYTE = 0x00000000
    PIPE_WAIT = 0x00000000
    ERROR_PIPE_CONNECTED = 535

    def CheckConnect(result, func, arguments):
        if result == 0:
            error = ctypes.get_last_error()
            if error != ERROR_PIPE_CONNECTED:
                raise ctypes.WinError(error)
        return result

    ConnectNamedPipe = kernel32.ConnectNamedPipe
    ConnectNamedPipe.restype = ctypes.wintypes.BOOL
    ConnectNamedPipe.argtypes = [ctypes.wintypes.HANDLE, ctypes.c_void_p]
    ConnectNamedPipe.errcheck = CheckConnect

    GetConsoleProcessList = kernel32.GetConsoleProcessList
    GetConsoleProcessList.restype = ctypes.wintypes.DWORD
    GetConsoleProcessList.argtypes = (ctypes.wintypes.LPDWORD, ctypes.wintypes.DWORD)


spawn_lock = threading.Lock()


try:
    import ubjson
    if os.name == 'nt':
        import psutil
    __all__ = ['AMSWorker', 'AMSWorkerResults', 'AMSWorkerError', 'AMSWorkerPool']
except ImportError:
    __all__ = []

from ...mol.molecule import Molecule
from ...core.settings import Settings
from ...core.errors import PlamsError, JobError, ResultsError
from ...tools.units import Units
from ...core.functions import config, log
from .ams import AMSJob
from .amspipeerror import *


def _restrict(func):
    """Decorator that wraps methods of |AMSWorkerResults| instances.

    This is used to replicate the behaviour of the full |AMSResults| object: Access to the values in an |AMSWorkerResults| instance will first check if the calculation leading to the results finish correctly and raise a |ResultsError| error exception if this is not the case. This behaviour can be modified with the ``config.ignore_failure`` setting.
    """
    @functools.wraps(func)
    def guardian(self, *args, **kwargs):
        if self.ok():
            return func(self, *args, **kwargs)
        else:
            if config.ignore_failure:
                log('WARNING: Trying to obtain results of a failed calculation {}'.format(self.name), 3)
                try:
                    ret = func(self, *args, **kwargs)
                except:
                    log('Obtaining results of {} failed. Returned value is None'.format(self.name), 3)
                    return None
                log('Obtaining results of {} successful. However, no guarantee that they make sense'.format(self.name), 3)
                return ret
            else:
                raise ResultsError('Using Results obtained from a failed calculation')
    return guardian



class AMSWorkerResults:
    """A specialized class encapsulating the results from calls to an |AMSWorker|.

    .. technical::

        AMSWorkerResults is *not* a subclass of |Results| or |AMSResults|. It does however implement some commonly used methods of the |AMSResults| class, so that results calculated by |AMSJob| and |AMSWorker| can be accessed in a uniform way.
    """

    def __init__(self, name, molecule, results, error=None):
        self._name           = name
        self._input_molecule = molecule
        self.error           = error
        self._results        = results
        if self._results is not None and 'xyzAtoms' in self._results:
            self._main_molecule = self._input_molecule.copy()
            self._main_molecule.from_array(self._results.pop('xyzAtoms') * Units.conversion_ratio('au', 'Angstrom'))
            if 'latticeVectors' in self._results:
                self._main_molecule.lattice = [ tuple(v) for v in self._results.pop('latticeVectors') * Units.conversion_ratio('au', 'Angstrom') ]
        else:
            self._main_molecule = self._input_molecule

    @property
    def name(self):
        """The name of a calculation.

        That is the name that was passed into the |AMSWorker| method when this |AMSWorkerResults| object was created. I can not be changed after the |AMSWorkerResults| instance has been created.
        """
        return self._name
    @name.setter
    def name(self, _):
        raise ResultsError('The name attribute of AMSWorkerResults may not be changed.')

    def ok(self):
        """Check if the calculation was successful. If not, the ``error`` attribute contains a corresponding exception.

        Users should check if the calculation was successful before using the other methods of the |AMSWorkerResults| instance, as using them might raise a |ResultsError| exception otherwise.
        """
        return self.error is None

    def get_errormsg(self):
        """Attempts to retreive a human readable error message from a crashed job. Returns ``None`` for jobs without errors."""
        if self.ok():
            return None
        else:
            lines = str(self.error).splitlines()
            if lines:
                for line in reversed(lines):
                    if 'ERROR: ' in line:
                        return line.partition('ERROR: ')[2]
                return lines[-1]
            else:
                return 'Could not determine error message. Please check the error.stdout and error.stderr manually.'

    @_restrict
    def get_energy(self, unit='au'):
        """Return the total energy, expressed in *unit*.
        """
        return self._results["energy"] * Units.conversion_ratio('au', unit)

    @_restrict
    def get_gradients(self, energy_unit='au', dist_unit='au'):
        """Return the nuclear gradients of the total energy, expressed in *energy_unit* / *dist_unit*.
        """
        return self._results["gradients"] * Units.conversion_ratio('au', energy_unit) / Units.conversion_ratio('au', dist_unit)

    @_restrict
    def get_stresstensor(self):
        """Return the clamped-ion stress tensor, expressed in atomic units.
        """
        return self._results["stressTensor"]

    @_restrict
    def get_hessian(self):
        """Return the Hessian matrix, i.e. the second derivative of the total energy with respect to the nuclear coordinates, expressed in atomic units.
        """
        return self._results["hessian"]

    @_restrict
    def get_elastictensor(self):
        """Return the elastic tensor, expressed in atomic units.
        """
        return self._results["elasticTensor"]

    @_restrict
    def get_charges(self):
        """Return the atomic charges, expressed in atomic units.
        """
        return self._results["charges"]

    @_restrict
    def get_dipolemoment(self):
        """Return the electric dipole moment, expressed in atomic units.
        """
        return self._results["dipoleMoment"]

    @_restrict
    def get_dipolegradients(self):
        """Return the nuclear gradients of the electric dipole moment, expressed in atomic units. This is a (3*numAtoms x 3) matrix.
        """
        return self._results["dipoleGradients"]

    def get_input_molecule(self):
        """Return a |Molecule| instance with the coordinates passed into the |AMSWorker|.

        Note that this method may also be used if the calculation producing this |AMSWorkerResults| object has failed, i.e. :meth:`ok` is ``False``."""
        return self._input_molecule

    @_restrict
    def get_main_molecule(self):
        """Return a |Molecule| instance with the final coordinates.
        """
        return self._main_molecule



class AMSWorkerError(PlamsError):
    """Error related to an AMSWorker process.

    The output from the failed worker process is stored in the ``stdout`` and ``stderr`` attributes.
    """

    def __init__(self, *args):
        super().__init__(*args)
        self.stdout = None
        self.stderr = None

    def __str__(self):
        msg = super().__str__()
        if self.stderr is not None:
            return "".join([msg, "\n"] + self.stderr)
        else:
            return msg

    def get_errormsg(self):
        lines = str(self).splitlines()
        if lines:
            for line in reversed(lines):
                if 'ERROR: ' in line:
                    return line.partition('ERROR: ')[2]
            return lines[-1]
        else:
            return 'Could not determine error message. Please check the error.stdout and error.stderr manually.'



_arg2setting = {}

for x in ("prev_results", "quiet"):
    _arg2setting[x] = ('amsworker', x)

for x in ("gradients", "stresstensor", "hessian", "elastictensor", "charges", "dipolemoment", "dipolegradients"):
    _arg2setting[x] = ('input', 'ams', 'properties', x)

for x in ("coordinatetype", "optimizelattice", "maxiterations", "pretendconverged", "calcpropertiesonlyifconverged"):
    _arg2setting[x] = ('input', 'ams', 'geometryoptimization', x)

for x in ("convenergy", "convgradients", "convstep", "convstressenergyperatom"):
    _arg2setting[x] = ('input', 'ams', 'geometryoptimization', 'convergence', x[4:])

_arg2setting['task'] = ('input', 'ams', 'task')
_arg2setting['usesymmetry'] = ('input', 'ams', 'usesymmetry')
_arg2setting['method'] = ('input', 'ams', 'geometryoptimization', 'method')

_setting2arg = {s: a for a, s in _arg2setting.items()}


class AMSWorker:
    """A class representing a running instance of the AMS driver as a worker process.

    Users need to supply a |Settings| instance representing the input of the AMS driver process (see :ref:`AMS_preparing_input`), but **not including** the ``Task`` keyword in the input (the ``input.ams.Task`` key in the |Settings| instance). The |Settings| instance should also not contain a system specification in the ``input.ams.System`` block, the ``input.ams.Properties`` block, or the ``input.ams.GeometryOptimization`` block. Often the settings of the AMS driver in worker mode will come down to just the engine block.

    The AMS driver will then start up as a worker, communicating with PLAMS via named pipes created in a temporary directory (determined by the *workerdir_root* and *workerdir_prefix* arguments). This temporary directory might also contain temporary files used by the worker process. Note that while an |AMSWorker| instance exists, the associated worker process can be assumed to be running and ready: If it crashes for some reason, it is automatically restarted.

    The recommended way to start an |AMSWorker| is as a context manager:

    .. code-block:: python

        with AMSWorker(settings) as worker:
            results = worker.SinglePoint('my_calculation', molecule)
        # clean up happens automatically when leaving the block

    If it is not possible to use the |AMSWorker| as a context manager, cleanup should be manually triggered by calling the :meth:`stop` method.
    """

    def __init__(self, settings, workerdir_root=TMPDIR, workerdir_prefix='amsworker', use_restart_cache=True, keep_crashed_workerdir=False):

        self.PyProtVersion = 1
        self.timeout = 2.0
        self.use_restart_cache = use_restart_cache

        # We'll initialize these communication related variables to None for now.
        # They will be overwritten when we actually start things up, but we do
        # not want them to be undefined for now, just in case of errors ...
        self.proc      = None
        self.callpipe  = None
        self.replypipe = None

        self.restart_cache = set()
        self.restart_cache_deleted = set()

        # Check if the settings we have are actually suitable for a PipeWorker.
        # They should not contain certain keywords and blocks.
        if 'ams' in settings.input:
            if 'Task' in settings.input.ams:
                raise JobError('Settings for AMSWorker should not contain a Task')
            if 'System' in settings.input.ams:
                raise JobError('Settings for AMSWorker should not contain a System block')
            if 'Properties' in settings.input.ams:
                raise JobError('Settings for AMSWorker should not contain the Properties block')
            if 'GeometryOptimization' in settings.input.ams:
                raise JobError('Settings for AMSWorker should not contain the GeometryOptimization block')

        # Make a copy of the Settings instance so we do not modify the outside world and fix the task to be "Pipe".
        self.settings = settings.copy()
        self.settings.input.ams.task = 'pipe'

        # Create the directory in which we will run the worker.
        self.wd_root   = workerdir_root
        self.wd_prefix = workerdir_prefix+'_'
        self.workerdir = tempfile.mkdtemp(dir=self.wd_root, prefix=self.wd_prefix)
        self._finalize = weakref.finalize(self, shutil.rmtree, self.workerdir)
        self.keep_wd   = keep_crashed_workerdir

        # Start the worker process.
        self._start_subprocess()


    def _start_subprocess(self):

        # We will use the standard PLAMS AMSJob class to prepare our input and runscript.
        amsjob = AMSJob(name='amsworker', settings=self.settings)
        with open(os.path.join(self.workerdir, 'amsworker.in'), 'w') as input_file:
            input_file.write(amsjob.get_input())
        with open(os.path.join(self.workerdir, 'amsworker.run'), 'w') as runscript:
            runscript.write(amsjob.get_runscript())
        del amsjob

        # Create the named pipes for communication.
        if os.name == 'nt':
            self._pipe_name = r"\\.\pipe\{}_amspipe".format(
                    os.path.abspath(self.workerdir).translate(
                        str.maketrans(r":\/", "___")
                    )
                )
            pipe = CreateNamedPipe(self._pipe_name, PIPE_ACCESS_DUPLEX,
                                   PIPE_TYPE_BYTE | PIPE_REJECT_REMOTE_CLIENTS | PIPE_WAIT,
                                   1,            # Maximum number of instances
                                   65536, 65536, # Output and input buffers in bytes
                                   60000,        # Timeout in ms (unused unless someone calls Wait)
                                   None)
        else:
            for filename in ["call_pipe", "reply_pipe"]:
                os.mkfifo(os.path.join(self.workerdir, filename))

        # Launch the worker process

        # spawn_lock is needed on Windows to workaround https://bugs.python.org/issue19575
        # Without it, multiple AMSWorkers in a pool would call Popen() simultaneously,
        # causing all the worker processes to inherit all the stdin/out/err file handles,
        # creating a huge mess and preventing the deletion of these files when a worker quits.
        # We can use spawn_lock on all platforms for simplicity without hurting performance
        # because most of the work is serialized in the kernel anyway.
        with spawn_lock:
            with open(os.path.join(self.workerdir, 'amsworker.in'), 'r') as amsinput, \
                 open(os.path.join(self.workerdir, 'ams.out'), 'w') as amsoutput, \
                 open(os.path.join(self.workerdir, 'ams.err'), 'w') as amserror:
                self.proc = subprocess.Popen(['sh', 'amsworker.run'], cwd=self.workerdir,
                                             stdout=amsoutput, stdin=amsinput, stderr=amserror,
                                             # Put all worker processes into a dedicated process group
                                             # to enable mass-killing in stop().
                                             start_new_session=(os.name == 'posix'),
                                             creationflags=(subprocess.CREATE_NEW_PROCESS_GROUP if os.name == 'nt' else 0)
                                             )

        # Start a dedicated watcher thread to rescue us in case the worker never opens its end of the pipes.
        self._stop_watcher = threading.Event()
        self._watcher = threading.Thread(target=self._startup_watcher, args=[self.workerdir], daemon=True)
        try:
            self._watcher.start()

            # This will block until either the worker is ready or the watcher steps in.
            if os.name == 'nt':
                ConnectNamedPipe(pipe, None)
                pipefd = msvcrt.open_osfhandle(pipe, 0)
                self.callpipe = os.fdopen(pipefd, 'r+b')
                self.replypipe = self.callpipe
            else:
                self.callpipe  = open(os.path.join(self.workerdir, 'call_pipe'), 'wb')
                self.replypipe = open(os.path.join(self.workerdir, 'reply_pipe'), 'rb')
        finally:
            # Both open()s are either done or have failed, we don't need the watcher thread anymore.
            self._stop_watcher.set()
            self._watcher.join()

        # Raise a nice error message if the worker failed to start. Otherwise, we'd get
        # a less descriptive error from the call to Hello below.
        try:
            if not self._check_process():
                raise AMSWorkerError('AMSWorker process did not start up correctly')

            # Now everything should be ready. Let's try saying Hello via the pipes ...
            self._call("Hello", {"version": self.PyProtVersion})
        except AMSWorkerError as exc:
            exc.stdout, exc.stderr = self.stop()
            raise


    def _startup_watcher(self, workerdir):
        while not self._stop_watcher.is_set():
            try:
                self.proc.wait(timeout=0.01)
                # self.proc has died and won't open its end of the pipes ...
                if not self._stop_watcher.is_set():
                    # ... but the main thread is still expecting someone to do it.
                    # Let's do it ourselves to unblock our main thread.
                    if os.name == 'nt':
                        with open(self._pipe_name, 'r+b'):
                            pass
                    else:
                        with open(os.path.join(workerdir, 'call_pipe'), 'rb'), \
                            open(os.path.join(workerdir, 'reply_pipe'), 'wb'):
                            # Nothing to do here, just close the pipes again.
                            pass
                return
            except subprocess.TimeoutExpired:
                # self.proc is still alive.
                pass


    def __enter__(self):
        return self


    if os.name == 'nt':
        def _find_worker_processes(self):
            # This is a convoluted workaround for the fact that the MSYS sh.exe on Windows likes to
            # launch commands as detached grandchildren, not direct children, so we can't track them
            # down just by following PPIDs. We thus have to resort to heuristics to find all processes
            # that we need to kill. It'd be best to use Win32 Job objects for this, but the "subprocess"
            # module doesn't let us add the created child to a Job early enough.

            # Get the PIDs of all processes sharing this console.
            bufsize = 1024
            while True:
                console_pids = (ctypes.wintypes.DWORD * bufsize)()
                n = GetConsoleProcessList(console_pids, bufsize)
                if n == 0:
                    raise ctypes.WinError(ctypes.get_last_error())
                elif n > bufsize:
                    bufsize *= 2
                else:
                    break

            # Convert PIDs to Process objects ASAP to minimize the potential for races with PID reuse.
            console_procs = []
            for pid in console_pids[0:n]:
                try:
                    console_procs.append(psutil.Process(pid))
                except psutil.Error:
                    # The process exited in the meantime or we can't access it, just skip it.
                    pass

            # Find all "(ba)sh.exe" processes on this console that are running in self.workerdir
            # and add them to worker_procs including all descendants.
            worker_procs = set()
            for proc in console_procs:
                if proc in worker_procs:
                    continue
                try:
                    if proc.exe().endswith("sh.exe") and proc.cwd() == self.workerdir:
                        worker_procs.add(proc)
                        worker_procs.update(proc.children(recursive=True))
                except psutil.Error:
                    pass

            return worker_procs


    def stop(self, keep_workerdir=False):
        """Stops the worker process and removes its working directory.

        This method should be called when the |AMSWorker| instance is not used as a context manager and the instance is no longer needed. Otherwise proper cleanup is not guaranteed to happen, the worker process might be left running and files might be left on disk.
        """

        stdout  = None
        stderr  = None

        # Tell the worker process to stop.
        if self.proc is not None:
            if self._check_process():
                try:
                    self._call("Exit")
                except AMSWorkerError:
                    # The process is likely exiting already.
                    if self.keep_wd:
                        keep_workerdir = True
                        print(f'AMSWorkerError encountered, will keep the workerdir in {self.workerdir}')

        # Tear down the pipes. Ignore OSError telling us the pipes are already broken.
        # This will also make the worker exit if it didn't get the Exit message above.
        if self.callpipe is not None:
            if not self.callpipe.closed:
                try:
                    self.callpipe.close()
                except OSError:
                    pass
            self.callpipe = None
        if self.replypipe is not None:
            if not self.replypipe.closed:
                try:
                    self.replypipe.close()
                except OSError:
                    pass
            self.replypipe = None

        # Now that the pipes are down, the worker should certainly be exiting.
        if self.proc is not None:
            try:
                self.proc.wait(timeout=self.timeout)
            except subprocess.TimeoutExpired:
                if os.name == 'nt':
                    worker_procs = self._find_worker_processes()
                    # Send Ctrl-Break to the entire process group under self.proc.
                    # Ctrl-C is less reliable in convincing processes to quit.
                    os.kill(self.proc.pid, signal.CTRL_BREAK_EVENT)
                    dead, alive = psutil.wait_procs(worker_procs, timeout=self.timeout)
                    for p in alive:
                        # Forcefully kill any descendant that is still running.
                        p.kill()
                    psutil.wait_procs(alive)
                else:
                    # Using SIGINT guarantees that the "sh" we're running as self.proc (and the "sh" running
                    # the start script etc.) will wait until its children have exited before exiting itself.
                    # The wait() thus doesn't return until all the descendants including ams.exe are gone.
                    # If we used SIGTERM instead, the "sh"s would exit immediately.
                    os.killpg(self.proc.pid, signal.SIGINT)
                    try:
                        self.proc.wait(timeout=self.timeout)
                    except subprocess.TimeoutExpired:
                        # This is guaranteed to stop everything, so we don't really have to wait for all processes.
                        os.killpg(self.proc.pid, signal.SIGKILL)

                self.proc.wait()
            self.proc = None
            with open(os.path.join(self.workerdir, 'ams.out'), 'r') as amsoutput:
                stdout = amsoutput.readlines()
            with open(os.path.join(self.workerdir, 'ams.err'), 'r') as amserror:
                stderr = amserror.readlines()

        # At this point the worker is down. We definitely don't have anything in the restart cache anymore ...
        self.restart_cache.clear()
        self.restart_cache_deleted.clear()

        if keep_workerdir:
            # Keep the current workerdir and move to a new one
            self._finalize.detach()
            self.workerdir = tempfile.mkdtemp(dir=self.wd_root, prefix=self.wd_prefix)
            self._finalize = weakref.finalize(self, shutil.rmtree, self.workerdir)
        else:
            # Remove the contents of the worker directory.
            for filename in os.listdir(self.workerdir):
                file_path = os.path.join(self.workerdir, filename)
                if os.path.isdir(file_path):
                    shutil.rmtree(file_path)
                else:
                    os.unlink(file_path)

        return (stdout, stderr)


    def __exit__(self, *args):
        self.stop()


    def _delete_from_restart_cache(self, name):
        if name in self.restart_cache:
            self.restart_cache.remove(name)
            self.restart_cache_deleted.add(name)


    def _prune_restart_cache(self):
        for name in list(self.restart_cache_deleted):
            self._call("DeleteResults", {"title": name})
        self.restart_cache_deleted.clear()


    @staticmethod
    def _supports_settings(s:Settings) -> bool:
        '''
        Check if a |Settings| object is supported by |AMSWorker|.
        '''

        try:
            args = AMSWorker._settings_to_args(s)
            return True
        except NotImplementedError:
            return False


    @staticmethod
    def _settings_to_args(s:Settings) -> Tuple[ str, Dict ]:
        '''
        Return a `tuple(TASK, **request_kwargs)` corresponding to a given settings object.

        Raises NotImplementedError if unsupported features are encountered.
        '''

        args = {}
        for key, val in s.flatten().items():
            kl = tuple(x.lower() if isinstance(x, str) else x for x in key )
            try:
                args[_setting2arg[kl]] = val
            except KeyError:
                raise NotImplementedError("Unexpected key {}".format(".".join(str(k) for k in key)))

        if "task" not in args:
            raise NotImplementedError("No Settings.input.ams.task found")
        elif args["task"].lower() not in ("singlepoint", "geometryoptimization"):
            raise NotImplementedError("Unexpected task {}".format(args["task"]))

        return args


    @staticmethod
    def _args_to_settings(**kwargs) -> Settings:
        s = Settings()
        for key, val in kwargs.items():
            s.set_nested(_arg2setting[key], val)
        return s


    def _solve_from_settings(self, name, molecule, settings):
        args = AMSWorker._settings_to_args(settings)
        return self._solve(name, molecule, **args)


    def _solve(self, name, molecule, task, prev_results=None, quiet=True,
               gradients=False, stresstensor=False, hessian=False, elastictensor=False,
               charges=False, dipolemoment=False, dipolegradients=False,
               method=None, coordinatetype=None, usesymmetry=None, optimizelattice=False,
               maxiterations=None, pretendconverged=None, calcpropertiesonlyifconverged=True,
               convenergy=None, convgradients=None, convstep=None, convstressenergyperatom=None):

        if self.use_restart_cache and name in self.restart_cache:
            raise JobError(f'Name "{name}" is already associated with results from the restart cache.')

        try:

            # This is a good opportunity to let the worker process know about all the results we no longer need ...
            self._prune_restart_cache()

            chemicalSystem = {}
            chemicalSystem['atomSymbols'] = np.asarray([atom.symbol for atom in molecule])
            chemicalSystem['coords'] = molecule.as_array() * Units.conversion_ratio('Angstrom','Bohr')
            if 'charge' in molecule.properties:
                chemicalSystem['totalCharge'] = float(molecule.properties.charge)
            else:
                chemicalSystem['totalCharge'] = 0.0
            self._call("SetSystem", chemicalSystem)
            if molecule.lattice:
                cell = np.asarray(molecule.lattice) * Units.conversion_ratio('Angstrom','Bohr')
                self._call("SetLattice", {"vectors": cell})
            else:
                self._call("SetLattice", {})

            args = {
                "request": { "title": str(name) },
                "keepResults": self.use_restart_cache,
            }
            if quiet: args["request"]["quiet"] = True
            if gradients: args["request"]["gradients"] = True
            if stresstensor: args["request"]["stressTensor"] = True
            if hessian: args["request"]["hessian"] = True
            if elastictensor: args["request"]["elasticTensor"] = True
            if charges: args["request"]["charges"] = True
            if dipolemoment: args["request"]["dipoleMoment"] = True
            if dipolegradients: args["request"]["dipoleGradients"] = True
            if self.use_restart_cache and prev_results is not None and prev_results.name in self.restart_cache:
                args["prevTitle"] = prev_results.name

            if task.lower() == 'geometryoptimization':
                if method is not None: args["method"] = str(method)
                if coordinatetype is not None: args["coordinateType"] = str(coordinatetype)
                if usesymmetry is not None: args["useSymmetry"] = bool(usesymmetry)
                if optimizelattice: args["optimizeLattice"] = True
                if maxiterations is not None: args["maxIterations"] = int(maxiterations)
                if pretendconverged: args["pretendConverged"] = True
                if not calcpropertiesonlyifconverged: args["calcPropertiesIfNotConverged"] = True
                if convenergy is not None: args["convEnergy"] = float(convenergy)
                if convgradients is not None: args["convGradients"] = float(convgradients)
                if convstep is not None: args["convStep"] = float(convstep)
                if convstressenergyperatom is not None: args["convStressEnergyPerAtom"] = float(convstressenergyperatom)
                results = self._call("Optimize", args)
            else:
                results = self._call("Solve", args)

            results = self._unflatten_arrays(results[0]['results'])
            results = AMSWorkerResults(name, molecule, results)

            if self.use_restart_cache:
                self.restart_cache.add(name)
                weakref.finalize(results, self._delete_from_restart_cache, name)

            return results

        except AMSPipeRuntimeError as exc:
            return AMSWorkerResults(name, molecule, None, exc)
        except AMSWorkerError as exc:
            # Something went wrong. Our worker process might also be down.
            # Let's reset everything to be safe ...
            exc.stdout, exc.stderr = self.stop()
            self._start_subprocess()
            # ... and return an AMSWorkerResults object indicating our failure.
            return AMSWorkerResults(name, molecule, None, exc)


    def SinglePoint(self, name, molecule, prev_results=None, quiet=True,
                    gradients=False, stresstensor=False, hessian=False, elastictensor=False,
                    charges=False, dipolemoment=False, dipolegradients=False):
        """Performs a single point calculation on the geometry given by the |Molecule| instance *molecule* and returns an instance of |AMSWorkerResults| containing the results.

        Every calculation should be given a *name*. Note that the name **must be unique** for this |AMSWorker| instance: One should not attempt to reuse calculation names with a given instance of |AMSWorker|.

        By default only the total energy is calculated but additional properties can be requested using the corresponding keyword arguments:

        - *gradients*: Calculate the nuclear gradients of the total energy.
        - *stresstensor*: Calculate the clamped-ion stress tensor. This should only be requested for periodic systems.
        - *hessian*: Calculate the Hessian matrix, i.e. the second derivative of the total energy with respect to the nuclear coordinates.
        - *elastictensor*: Calculate the elastic tensor. This should only be requested for periodic systems.
        - *charges*: Calculate atomic charges.
        - *dipolemoment*: Calculate the electric dipole moment. This should only be requested for non-periodic systems.
        - *dipolemoment*: Calculate the nuclear gradients of the electric dipole moment. This should only be requested for non-periodic systems.

        Users can pass an instance of a previously obtained |AMSWorkerResults| as the *prev_results* keyword argument. This can trigger a restart from previous results in the worker process, the details of which depend on the used computational engine: For example, a DFT based engine might restart from the electronic density obtained in an earlier calculation on a similar geometry. This is often useful to speed up series of sequentially dependent calculations:

        .. code-block:: python

            mol = Molecule('some/system.xyz')
            with AMSWorker(sett) as worker:
                last_results = None
                do i in range(num_steps):
                    results = worker.SinglePoint(f'step{i}', mol, prev_results=last_results, gradients=True)
                    # modify the geometry of mol using results.get_gradients()
                    last_results = results

        Note that the restarting is disabled if the |AMSWorker| instance was created with ``use_restart_cache=False``. It is still permitted to pass previous |AMSResults| instances as the *prev_results* argument, but no restarting will happen.

        The *quiet* keyword can be used to obtain more output from the worker process. Note that the output of the worker process is not printed to the standard output but instead ends up in the ``ams.out`` file in the temporary working directory of the |AMSWorker| instance. This is mainly useful for debugging.
        """
        args = locals()
        del args['self']
        del args['name']
        del args['molecule']
        s = self._args_to_settings(**args)
        s.input.ams.task = 'singlepoint'
        return self._solve_from_settings(name, molecule, s)


    def GeometryOptimization(self, name, molecule, prev_results=None, quiet=True,
                             gradients=False, stresstensor=False, hessian=False, elastictensor=False,
                             charges=False, dipolemoment=False, dipolegradients=False,
                             method=None, coordinatetype=None, usesymmetry=None, optimizelattice=False,
                             maxiterations=None, pretendconverged=None, calcpropertiesonlyifconverged=True,
                             convenergy=None, convgradients=None, convstep=None, convstressenergyperatom=None):
        """Performs a geometry optimization on the |Molecule| instance *molecule* and returns an instance of |AMSWorkerResults| containing the results from the optimized geometry.

        The geometry optimizer can be controlled using the following keyword arguments:

        - *method*: String identifier of a particular optimization algorithm.
        - *coordinatetype*: Select a particular kind of optimization coordinates.
        - *usesymmetry*: Enable the use of symmetry when applicable.
        - *optimizelattice*: Optimize the lattice vectors together with atomic positions.
        - *maxiterations*: Maximum number of iterations allowed.
        - *pretendconverged*: If set to true, non converged geometry optimizations will be considered successful.
        - *calcpropertiesonlyifconverged*: Calculate properties (e.g. the Hessian) only if the optimization converged.
        - *convenergy*: Convergence criterion for the energy (in Hartree).
        - *convgradients*: Convergence criterion for the gradients (in Hartree/Bohr).
        - *convstep*: Convergence criterion for displacements (in Bohr).
        - *convstressenergyperatom*: Convergence criterion for the stress energy per atom (in Hartree).
        """
        args = locals()
        del args['self']
        del args['name']
        del args['molecule']
        s = self._args_to_settings(**args)
        s.input.ams.task = 'geometryoptimization'
        return self._solve_from_settings(name, molecule, s)


    def ParseInput(self, program_name, text_input):
        try:
            reply = self._call("ParseInput", {"programName": program_name, "textInput": text_input})
            json_input = reply[0]['parsedInput']['jsonInput']
            return json_input
        except AMSWorkerError as exc:
            # This failed badly, also the worker is likely down. Let's grab some info, restart it ...
            exc.stdout, exc.stderr = self.stop()
            self._start_subprocess()
            # ... and then reraise the exception for the caller.
            raise


    def _check_process(self) :
        if self.proc is not None:
            status = self.proc.poll()
            return status is None
        else:
            return False


    def _flatten_arrays(self, d):
        out = {}
        for key, val in d.items():
            if (isinstance(val, collections.abc.Sequence) or isinstance(val, np.ndarray)) and not isinstance(val, str):
                array = np.asarray(val)
                out[key] = array.flatten()
                out[key + "_dim_"] = array.shape[::-1]
            elif isinstance(val, collections.abc.Mapping):
                out[key] = self._flatten_arrays(val)
            else:
                out[key] = val
        return out


    def _unflatten_arrays(self, d):
        out = {}
        for key, val in d.items():
            if key + "_dim_" in d:
                out[key] = np.asarray(val).reshape(d[key + "_dim_"][::-1])
            elif key.endswith("_dim_"):
                pass
            elif isinstance(val, collections.abc.Mapping):
                out[key] = self._unflatten_arrays(val)
            else:
                out[key] = val
        return out

    def _read_exactly(self, pipe, n):
        buf = pipe.read(n)
        if len(buf) == n:
            return buf
        else:
            raise EOFError("Message truncated")

    def _call(self, method, args={}):
        msg = ubjson.dumpb({method: self._flatten_arrays(args)})
        msglen = struct.pack("=i", len(msg))
        try:
            self.callpipe.write(msglen + msg)
            if method.startswith("Set"):
                return None
            self.callpipe.flush()
        except OSError as exc:
            raise AMSWorkerError('Error while sending a message') from exc
        if method == "Exit":
            return None

        results = []
        while True:
            try:
                msgbuf = self._read_exactly(self.replypipe, 4)
                msglen = struct.unpack("=i", msgbuf)[0]
                msgbuf = self._read_exactly(self.replypipe, msglen)
            except EOFError as exc:
                raise AMSWorkerError("Error while trying to read a reply") from exc

            try:
                msg = ubjson.loadb(msgbuf)
            except Exception as exc:
                raise AMSWorkerError("Error while decoding a reply") from exc

            if "return" in msg:
                ret = msg["return"]
                if ret["status"] == 0:
                    return results
                else:
                    raise AMSPipeError.from_message(ret)
            else:
                results.append(msg)



class AMSWorkerPool:
    """A class representing a pool of AMS worker processes.

    All workers of the pool are initialized with the same |Settings| instance, see the |AMSWorker| constructor for details.

    The number of spawned workers is determined by the *num_workers* argument. For optimal performance on many small jobs it is recommended to spawn a number of workers equal to the number of physical CPU cores of the machine the calculation is running on, and to let every worker instance run serially:

    .. code-block:: python

        import psutil

        molecules = read_molecules('folder/with/xyz/files')

        sett = Settings()
        # ... more settings ...
        sett.runscript.nproc = 1 # <-- every worker itself is serial (aka export NSCM=1)

        with AMSWorkerPool(sett, psutil.cpu_count(logical=False)) as pool:
            results = pool.SinglePoints([ (name, molecules[name]) for name in sorted(molecules) ])

    As with the underlying |AMSWorker| class, the location of the temporary directories can be changed with the *workerdir_root* and *workerdir_prefix* arguments.

    It is recommended to use the |AMSWorkerPool| as a context manager in order to ensure that cleanup happens automatically. If it is not possible to use the |AMSWorkerPool| as a context manager, cleanup should be manually triggered by calling the :meth:`stop` method.

    """

    def __init__(self, settings, num_workers, workerdir_root=TMPDIR, workerdir_prefix='awp', keep_crashed_workerdir=False):

        self.workers = num_workers * [None]
        if num_workers == 1:
            # Do all the work in the main thread
            AMSWorkerPool._spawn_worker(self.workers, settings, 0, workerdir_root, workerdir_prefix, keep_crashed_workerdir)
        else:
            # Spawn all workers from separate threads to overlap the ams.exe startup latency
            threads = [ threading.Thread(target=AMSWorkerPool._spawn_worker, args=(self.workers, settings, i, workerdir_root, workerdir_prefix, keep_crashed_workerdir))
                        for i in range(num_workers) ]
            for t in threads: t.start()
            for t in threads: t.join()
            if None in self.workers:
                raise AMSWorkerError('Some AMSWorkers in the pool failed to start')


    @staticmethod
    def _spawn_worker(workers, settings, i, wdr, wdp, keep_crashed_workerdir):
        workers[i] = AMSWorker(settings, workerdir_root=wdr, workerdir_prefix=f'{wdp}_{i}', use_restart_cache=False, keep_crashed_workerdir=keep_crashed_workerdir)


    def __enter__(self):
        return self


    def _solve_from_settings(self, items):
        """Request to pool to execute calculations for all items in the iterable *items*. Returns a list of |AMSWorkerResults| objects.

        The *items* argument is expected to be an iterable of 3-tuples ``(name, molecule, settings)``, which are passed on to the the :meth:`_solve_from_settings <AMSWorker._solve_from_settings>` method of the pool's |AMSWorker| instances.
        """


        if len(self.workers) == 1:

            # Do all the work in the main thread
            results = [ self.workers[0]._solve_from_settings(name, mol, settings) for name, mol, settings in items ]

        else:

            # Build a queue of things to do and spawn threads that grab from from the queue in parallel
            results = [None]*len(items)
            q = queue.Queue()

            threads = [ threading.Thread(target=AMSWorkerPool._execute_queue, args=(self.workers[i], q, results)) for i in range(len(self.workers)) ]
            for t in threads: t.start()

            for i, item in enumerate(items):
                if len(item) == 3:
                    name, mol, settings = item
                else:
                    raise JobError('AMSWorkerPool._solve_from_settings expects a list containing only 3-tuples (name, molecule, settings).')
                q.put((i, name, mol, settings))

            for t in threads: q.put(None) # signal for the thread to end
            for t in threads: t.join()

        return results


    def _prep_solve_from_settings(self, method, items):

        solve_items = []
        for item in items:
            if len(item) == 2:
                name, mol = item
                kwargs = {}
            elif len(item) == 3:
                name, mol, kwargs = item
            else:
                raise JobError(f'AMSWorkerPool.{method}s expects a list containing only 2-tuples (name, molecule) and/or 3-tuples (name, molecule, kwargs).')
            s = AMSWorker._args_to_settings(**kwargs)
            s.input.ams.task = method.lower()

            solve_items.append((name, mol, s))

        return solve_items


    def SinglePoints(self, items):
        """Request to pool to execute single point calculations for all items in the iterable *items*. Returns a list of |AMSWorkerResults| objects.

        The *items* argument is expected to be an iterable of 2-tuples ``(name, molecule)`` and/or 3-tuples ``(name, molecule, kwargs)``, which are passed on to the :meth:`SinglePoint <AMSWorker.SinglePoint>` method of the pool's |AMSWorker| instances. (Here ``kwargs`` is a dictionary containing the optional keyword arguments and their values for this method.)

        As an example, the following call would do single point calculations with gradients and (only for periodic systems) stress tensors for all |Molecule| instances in the dictionary ``molecules``.

        .. code-block:: python

            results = pool.SinglePoint([ (name, molecules[name], {
                                             "gradients": True,
                                             "stresstensor": len(molecules[name].lattice) != 0
                                          }) for name in sorted(molecules) ])
        """
        solve_items = self._prep_solve_from_settings('SinglePoint', items)
        return self._solve_from_settings(solve_items)



    def GeometryOptimizations(self, items):
        """Request to pool to execute geometry optimizations for all items in the iterable *items*. Returns a list of |AMSWorkerResults| objects for the optimized geometries.

        The *items* argument is expected to be an iterable of 2-tuples ``(name, molecule)`` and/or 3-tuples ``(name, molecule, kwargs)``, which are passed on to the :meth:`GeometryOptimization <AMSWorker.GeometryOptimization>` method of the pool's |AMSWorker| instances. (Here ``kwargs`` is a dictionary containing the optional keyword arguments and their values for this method.)
        """
        solve_items = self._prep_solve_from_settings('GeometryOptimization', items)
        return self._solve_from_settings(solve_items)



    @staticmethod
    def _execute_queue(worker, q, results):
        while True:
            item = q.get()
            try:
                if item is None:
                    break
                i, name, mol, settings = item
                results[i] = worker._solve_from_settings(name, mol, settings)
            finally:
                q.task_done()


    def stop(self):
        """Stops the all worker processes and removes their working directories.

        This method should be called when the |AMSWorkerPool| instance is not used as a context manager and the instance is no longer needed. Otherwise proper cleanup is not guaranteed to happen, worker processes might be left running and files might be left on disk.
        """
        for worker in self.workers:
            worker.stop()


    def __exit__(self, *args):
        self.stop()



from functools import reduce

from ...core.basejob  import SingleJob
from ...core.settings import Settings


__all__ = ['GamessJob']


# ======================<>===========================


class GamessJob(SingleJob):
    """
    A class representing a single computational job with
    `GAMESS-US <http://www.msg.ameslab.gov/gamess/>`.
    A typical gamess input looks like:

      $contrl scftyp=rohf mult=2 cctyp=eom-ccsd runtyp=energy ispher=1 $end
      $system mwords=3 $end
      $basis  gbasis=accd $end
      $guess  guess=huckel $end
      $ccinp  ccprp=.false. $end
      $eominp nstate(1)=1,0,0,0 iroot(1)=1,1 ccprpe=.false.
              minit=1 noact=3 nuact=5 $end
      $data
     NH2...aug-cc-pVDZ basis...excited state @ g.s. MP2 geometry
     Cnv 2

     NITROGEN    7.0   0.0   0.0           -0.0370366120
     HYDROGEN    1.0   0.0  -0.8055298238  -0.6814816939
      $end
    Using Plams Settings it is written like:

    j = GamessJob()

    # data
    header = "NH2...aug-cc-pVDZ basis...excited state @ g.s. MP2 geometry\n"
    symmetry = "Cnv 2\n\n"
    n = "NITROGEN    7.0   0.0   0.0           -0.0370366120\n"
    h = "HYDROGEN    1.0   0.0  -0.8055298238  -0.6814816939"

    j.settings.input.data = header + symmetry + n + h
    # basis
    j.settings.input.basis.gbasis = 'accd'
    # guess
    j.settings.input.guess.guess = 'huckel'
    # ccinp
    j.settings.input.ccinp.ccprp = '.false.'
    # eominp
    j.settings.input.eominp['nstate(1)'] = '1,0,0,0'
    j.settings.input.eominp['iroot(1)'] = '1,1'
    j.settings.input.eominp.ccprpe = '.false.'
    j.settings.input.eominp.minit = 1
    j.settings.input.eominp.noact = 3
    j.settings.input.eominp.nuact = 5
    # control
    j.settings.input.contrl.scftyp = 'rhf'
    j.settings.input.contrl.runtype = 'optimize'
    j.settings.input.contrl.pp = 'mcp'
    j.settings.input.contrl.ispher = 1
    j.settings.input.contrl.coord = 'zmt'
    j.settings.input.contrl.nzvar = 6
    """
    _filenames = {'inp': '$JN.inp', 'run': ' $JN.run', 'out': '$JN.out',
                  'err': '$JN.err'}

    def get_input(self):
        """
        Transform all contents of ``input`` branch of ``settings`` into string
        with blocks, subblocks, keys and values.
        """
        def parse(key, value, indent=''):
            ret = ''
            if isinstance(value, Settings):
                ret += ' ${} '.format(key)
                for el in value:
                    ret += ' {}={} '.format(el, value[el])
                ret += '$end\n'
            else:
                ret += ' ${}\n{}\n $end\n'.format(key, value)

            return ret

        inp = [parse(item, self.settings.input[item])
               for item in self.settings.input]

        return self.print_molecule() + reduce(lambda x, y: x + y, inp)

    def print_molecule(self):
        """
        pretty print a molecule in the GAMESS format.
        """
        mol = self.molecule
        if mol:
            # Read Symmetry from properties otherwise use C1
            if 'symmetry' in mol.properties.keys():
                sym = mol.properties['symmetry']
            else:
                sym = 'C1'
            if sym != 'C1':
                sym += '\n'
            ret = ' $data\ntitle\n{}\n'.format(sym)
            for at in mol.atoms:
                ret += "{} {}   {}\n".format(at.symbol, at.atnum,
                                             at.str(symbol=False, space=14, decimal=10))
            ret += " $end\n"
            return ret
        else:
            return '\n'

    def get_runscript(self):
        """
        Run Gamess Using rungms.
        """
        return  'rungms {} > {}'.format(self._filename('inp'), self._filename('out'))

    def check(self):
        """
        Look for the normal termination signal in Cp2k output
        """
        s = self.results.grep_output("EXECUTION OF GAMESS TERMINATED NORMALLY")
        return len(s) > 0
